import argparse
import contextlib
import dataclasses
import functools
import io
import json
import logging
import typing as t
from unittest.mock import patch

import yaml

from cratedb_toolkit.exception import CroudException

if t.TYPE_CHECKING:
    from croud.parser import Argument


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class CroudCall:
    fun: t.Callable
    specs: t.List["Argument"]
    arguments: t.List[str]


class CroudWrapper:
    def __init__(self, call: CroudCall, output_format: str = None, decode_output: bool = True):
        """
        format: One of table,wide,json,yaml
        """
        output_format = output_format or "json"
        self.call = call
        self.output_format = output_format
        self.decode_output = decode_output

    def invoke(self) -> t.Any:
        """
        Invoke croud function, and enrich response with explanation.
        """
        try:
            return self.invoke_real()
        except CroudException as ex:
            message = str(ex)
            logger.error(f"The operation failed: {ex}")
            explanation = None
            if "Unauthorized" in message:
                # FIXME: Complete error message.
                explanation = (
                    "Please make sure you logged in successfully with `croud login`, and `croud me` works well."
                )
            if "Resource not found" in message:
                explanation = (
                    "Apparently, the addressed cluster resource was not found. "
                    "Please validate the `--cluster-id` option, or the `CRATEDB_CLOUD_CLUSTER_ID` environment variable "
                    "used by CrateDB Toolkit."
                )
            if "Another cluster operation is currently in progress" in message:
                explanation = (
                    "The cluster deployment is probably still in progress. "
                    "Please wait until the cluster has been fully deployed."
                )
            if explanation:
                logger.warning(explanation)
            raise
        except Exception:
            raise

    def invoke_real(self) -> t.Any:
        """
        Invoke croud function, decode response from JSON, and return data.
        """
        from croud.parser import add_default_args, create_parser

        # Create argument parser.
        spec: t.Dict[str, t.Any] = {
            "help": None,
            "extra_args": [],
        }
        parser = create_parser(spec)
        add_default_args(parser, omit=set())

        # Add command-specific arguments to parser.
        for arg_spec in self.call.specs:
            try:
                arg_spec.add_to_parser(parser)
            except argparse.ArgumentError:
                pass

        # Decode arguments.
        args_in = self.call.arguments + [f"--output-fmt={self.output_format}"]
        args = parser.parse_args(args_in)

        # Invoke croud function, with capturing.
        data = self.invoke_capturing(self.call.fun, args)

        if self.decode_output:
            if self.output_format == "json":
                return json.loads(data)
            elif self.output_format == "yaml":
                return yaml.safe_load(data)
        return data

    def invoke_capturing(self, fun: t.Callable, *args: t.List[t.Any], **kwargs: t.Dict[t.Any, t.Any]) -> str:
        """
        Invoke croud function, and return captured stdout.
        """
        croud_fun = functools.partial(fun, *args, **kwargs)
        buffer = io.StringIO()
        with contextlib.redirect_stdout(buffer):
            self.run_croud_fun(croud_fun)
        buffer.seek(0)
        return buffer.read()

    @staticmethod
    def run_croud_fun(fun: t.Callable, with_exceptions: bool = True):
        """
        Wrapper function to call into `croud`, for catching and converging error messages.

        The `croud` package, being conceived as a CLI program, uses `print` to output
        errors and other messages to stderr.

        This wrapper function catches them, emits them to the logger with an additional
        explanation, and eventually raises a `RuntimeException` on significant errors.
        That usually means that a CLI program will exit with an error code != 0.
        """

        # Central interceptor function for `print_*` functions.
        # Obtain level name as string as first positional argument.
        def print_fun(levelname: str, *args, **kwargs):
            level = get_sane_log_level(levelname)
            message = str(args[0])
            logger.log(level, message)
            if with_exceptions and level >= logging.ERROR:
                raise CroudException(message)

        # Patch all `print_*` functions, and invoke_foo workhorse function.
        # https://stackoverflow.com/a/46481946
        levels = ["debug", "info", "warning", "error", "success"]
        with contextlib.ExitStack() as stack:
            for level in levels:
                p = patch(f"croud.printer.print_{level}", functools.partial(print_fun, level))
                stack.enter_context(p)
            """
            TODO: When aiming to disable wait-for-completion.
                  p = patch(f"croud.clusters.commands._wait_for_completed_operation")
                  stack.enter_context(p)
            """
            return fun()


def get_sane_log_level(src) -> int:
    """
    https://stackoverflow.com/questions/18392804/python-logging-determine-level-number-from-name
    """
    level = logging.getLevelName(str(src).upper())
    if not isinstance(level, int):
        level = logging.NOTSET
    return level


def get_croud_output_formats() -> t.List[str]:
    """
    Inquire the output formats `croud` understands.
    """
    from croud.config.schemas import OUTPUT_FORMATS

    return OUTPUT_FORMATS



import logging

from influxio.core import copy

logger = logging.getLogger(__name__)


def influxdb_copy(source_url, target_url, progress: bool = False):
    """
    Synopsis
    --------
    export CRATEDB_SQLALCHEMY_URL=crate://crate@localhost:4200/testdrive/demo
    ctk load table influxdb2://example:token@localhost:8086/testdrive/demo
    """
    logger.info("Running InfluxDB copy")
    copy(source_url, target_url, progress=progress)
    return True



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# This is for Python 3.7 and 3.8 to support generic types
# like `dict` instead of `typing.Dict
from __future__ import annotations

import logging
import os
from typing import Optional

from testcontainers.core.config import MAX_TRIES
from testcontainers.core.generic import DbContainer
from testcontainers.core.waiting_utils import wait_container_is_ready, wait_for_logs

from cratedb_toolkit.testing.testcontainers.util import KeepaliveContainer, asbool
from cratedb_toolkit.util import DatabaseAdapter

logger = logging.getLogger(__name__)


class CrateDBContainer(KeepaliveContainer, DbContainer):
    """
    CrateDB database container.

    Example:

        The example spins up a CrateDB database and connects to it using
        SQLAlchemy and its Python driver.

        .. doctest::

            >>> from tests.testcontainers.cratedb import CrateDBContainer
            >>> import sqlalchemy

            >>> cratedb_container = CrateDBContainer("crate:5.2.3")
            >>> cratedb_container.start()
            >>> with cratedb_container as cratedb:
            ...     engine = sqlalchemy.create_engine(cratedb.get_connection_url())
            ...     with engine.begin() as connection:
            ...         result = connection.execute(sqlalchemy.text("select version()"))
            ...         version, = result.fetchone()
            >>> version
            'CrateDB 5.2.3...'
    """

    CRATEDB_USER = os.environ.get("CRATEDB_USER", "crate")
    CRATEDB_PASSWORD = os.environ.get("CRATEDB_PASSWORD", "")
    CRATEDB_DB = os.environ.get("CRATEDB_DB", "doc")
    KEEPALIVE = asbool(os.environ.get("CRATEDB_KEEPALIVE", os.environ.get("TC_KEEPALIVE", False)))
    CMD_OPTS = {
        "discovery.type": "single-node",
        "node.attr.storage": "hot",
        "path.repo": "/tmp/snapshots",
    }

    def __init__(
        self,
        image: str = "crate/crate:nightly",
        ports: Optional[dict] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        dbname: Optional[str] = None,
        cmd_opts: Optional[dict] = None,
        **kwargs,
    ) -> None:
        """
        :param image: docker hub image path with optional tag
        :param ports: optional dict that maps a port inside the container to a port on the host machine;
                      `None` as a map value generates a random port;
                      Dicts are ordered. By convention the first key-val pair is designated to the HTTP interface.
                      Example: {4200: None, 5432: 15432} - port 4200 inside the container will be mapped
                      to a random port on the host, internal port 5432 for PSQL interface will be mapped
                      to the 15432 port on the host.
        :param user:  optional username to access the DB; if None, try `CRATEDB_USER` environment variable
        :param password: optional password to access the DB; if None, try `CRATEDB_PASSWORD` environment variable
        :param dbname: optional database name to access the DB; if None, try `CRATEDB_DB` environment variable
        :param cmd_opts: an optional dict with CLI arguments to be passed to the DB entrypoint inside the container
        :param kwargs: misc keyword arguments
        """
        super().__init__(image=image, **kwargs)

        self._name = "testcontainers-cratedb"

        cmd_opts = cmd_opts or {}
        self._command = self._build_cmd({**self.CMD_OPTS, **cmd_opts})

        self.CRATEDB_USER = user or self.CRATEDB_USER
        self.CRATEDB_PASSWORD = password or self.CRATEDB_PASSWORD
        self.CRATEDB_DB = dbname or self.CRATEDB_DB

        self.port_mapping = ports if ports else {4200: None}
        self.port_to_expose, _ = list(self.port_mapping.items())[0]

    @staticmethod
    def _build_cmd(opts: dict) -> str:
        """
        Return a string with command options concatenated and optimised for ES5 use
        """
        cmd = []
        for key, val in opts.items():
            if isinstance(val, bool):
                val = str(val).lower()
            cmd.append(f"-C{key}={val}")
        return " ".join(cmd)

    def _configure_ports(self) -> None:
        """
        Bind all the ports exposed inside the container to the same port on the host
        """
        # If host_port is `None`, a random port to be generated
        for container_port, host_port in self.port_mapping.items():
            self.with_bind_ports(container=container_port, host=host_port)

    def _configure_credentials(self) -> None:
        self.with_env("CRATEDB_USER", self.CRATEDB_USER)
        self.with_env("CRATEDB_PASSWORD", self.CRATEDB_PASSWORD)
        self.with_env("CRATEDB_DB", self.CRATEDB_DB)

    def _configure(self) -> None:
        self._configure_ports()
        self._configure_credentials()

    def get_connection_url(self, dialect: str = "crate", host: Optional[str] = None) -> str:
        """
        Return a connection URL to the DB

        :param host: optional string
        :param dialect: a string with the dialect name to generate a DB URI
        :return: string containing a connection URL to te DB
        """
        # TODO: When using `db_name=self.CRATEDB_DB`:
        #       Connection.__init__() got an unexpected keyword argument 'database'
        return super()._create_connection_url(
            dialect=dialect,
            username=self.CRATEDB_USER,
            password=self.CRATEDB_PASSWORD,
            host=host,
            port=self.port_to_expose,
        )

    @wait_container_is_ready()
    def _connect(self):
        # TODO: Better use a network connectivity health check?
        #       In `testcontainers-java`, there is the `HttpWaitStrategy`.
        # TODO: Provide a client instance.
        wait_for_logs(self, predicate="o.e.n.Node.*started", timeout=MAX_TRIES)


class CrateDBTestAdapter:
    """
    A little helper wrapping Testcontainer's `CrateDBContainer` and
    CrateDB Toolkit's `DatabaseAdapter`, agnostic of the test framework.
    """

    def __init__(self, crate_version: str = "nightly", **kwargs):
        self.cratedb: Optional[CrateDBContainer] = None
        self.database: Optional[DatabaseAdapter] = None
        self.image: str = "crate/crate:{}".format(crate_version)

    def start(self, **kwargs):
        """
        Start testcontainer, used for tests set up
        """
        self.cratedb = CrateDBContainer(image=self.image, **kwargs)
        self.cratedb.start()
        self.database = DatabaseAdapter(dburi=self.get_connection_url())

    def stop(self):
        """
        Stop testcontainer, used for tests tear down
        """
        if self.cratedb:
            self.cratedb.stop()

    def reset(self, tables: Optional[list] = None):
        """
        Drop tables from the given list, used for tests set up or tear down
        """
        if tables and self.database:
            for reset_table in tables:
                self.database.connection.exec_driver_sql(f"DROP TABLE IF EXISTS {reset_table};")

    def get_connection_url(self, *args, **kwargs):
        """
        Return a URL for SQLAlchemy DB engine
        """
        if self.cratedb:
            return self.cratedb.get_connection_url(*args, **kwargs)
        return None

    def get_http_url(self, **kwargs):
        """
        Return a URL for CrateDB's HTTP endpoint
        """
        return self.get_connection_url(dialect="http", **kwargs)

    @property
    def http_url(self):
        """
        Return a URL for CrateDB's HTTP endpoint.

        Used to stay backward compatible with the downstream code.
        """
        return self.get_http_url()



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import logging
from importlib.resources import read_text

from cratedb_toolkit.retention.model import JobSettings
from cratedb_toolkit.util.database import run_sql

logger = logging.getLogger(__name__)


def setup_schema(settings: JobSettings):
    """
    Set up the retention policy table schema.

    TODO: Refactor to `store` module.
    """

    logger.info(
        f"Installing retention policy bookkeeping table at "
        f"database '{settings.database.safe}', table {settings.policy_table}"
    )

    # Read SQL DDL statement.
    sql = read_text("cratedb_toolkit.retention.setup", "schema.sql")

    tplvars = settings.to_dict()
    sql = sql.format_map(tplvars)

    if settings.dry_run:
        logger.info(f"Pretending to execute SQL statement:\n{sql}")
        return

    # Materialize table schema.
    run_sql(settings.database.dburi, sql)



import dataclasses
import typing as t

import requests
import sqlparse

from cratedb_toolkit.util import DatabaseAdapter

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore[assignment]


@dataclasses.dataclass
class Dataset:
    """
    Wrapper around a dataset.
    """

    reference: str
    file: t.Optional[str] = None
    ddl: t.Optional[str] = None
    title: t.Optional[str] = None
    documentation: t.Optional[str] = None
    data_url: t.Optional[str] = None
    init_url: t.Optional[str] = None
    init_includes_loading: t.Optional[bool] = False

    # Native datasets wrapper handle, e.g. `datasets`
    # https://github.com/huggingface/datasets
    handle: t.Optional[t.Any] = None

    def acquire(self):
        if self.reference:
            if self.reference.startswith("huggingface://"):
                from datasets import load_dataset as load_dataset_huggingface

                upstream_name = self.reference.replace("huggingface://", "")
                self.handle = load_dataset_huggingface(upstream_name)
            if self.reference.startswith("kaggle://"):
                from cratedb_toolkit.datasets.util import load_dataset_kaggle

                # TODO: Use an URL library to split off the file fragment part.
                # kaggle://guillemservera/global-daily-climate-data/daily_weather.parquet
                # =>
                # kaggle://guillemservera/global-daily-climate-data, daily_weather.parquet
                upstream_name = self.reference.replace("kaggle://", "")
                parts = upstream_name.rsplit("/", 1)
                upstream_name = parts[0]
                self.file = parts[1]

                # FIXME: Don't use `DOWNLOAD` as folder. Use platformdirs.
                load_dataset_kaggle(dataset=upstream_name, path="DOWNLOAD")

    def dbtable(self, dburi: str, table: str):
        return DatasetToDatabaseTableAdapter(dataset=self, dburi=dburi, table=table)


@dataclasses.dataclass
class DatasetToDatabaseTableAdapter:
    """
    Run a dataset into a database table.
    """

    dataset: Dataset
    dburi: str
    table: str

    def __post_init__(self):
        self.init_sql = None
        self.db = DatabaseAdapter(dburi=self.dburi)

    def create(
        self,
        if_exists: Literal["append", "noop", "replace"] = "noop",
        drop: bool = False,
    ):
        """
        Load dataset into CrateDB, or prepare loading, using SQL.
        """

        self.dataset.acquire()
        if self.dataset.init_url:
            self.init_sql = requests.get(self.dataset.init_url, timeout=5).text
        elif self.dataset.ddl:
            self.init_sql = self.dataset.ddl

        if self.init_sql:
            self.init_sql = self.init_sql.format(table=self.table)

            if drop:
                self.db.drop_table(self.table)

            if if_exists == "replace":
                self.db.prune_table(self.table, errors="ignore")

            if not self.dataset.init_includes_loading:
                self.run_sql(self.init_sql)

    def load(
        self,
        if_exists: Literal["append", "noop", "replace"] = "noop",
        drop: bool = False,
    ):
        self.create(if_exists=if_exists, drop=drop)

        cardinality = self.db.count_records(self.table, errors="ignore")
        has_data = cardinality > 0

        if if_exists == "noop" and has_data:
            return

        if self.init_sql is None:
            raise ValueError("SQL for loading data is missing")
        self.run_sql(self.init_sql)

    def run_sql(self, sql: str):
        for statement in sqlparse.parse(sql):
            self.db.run_sql(str(statement))


class DatasetRegistry:
    """
    Manage multiple datasets, and provide discovery.
    """

    def __init__(self):
        self.datasets: t.List[Dataset] = []

    def add(self, dataset: Dataset):
        self.datasets.append(dataset)

    def find(self, reference: str) -> Dataset:
        for dataset in self.datasets:
            if dataset.reference == reference:
                return dataset
        raise KeyError(f"Dataset not found: {reference}")



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import abc
import dataclasses
import logging
import os
import typing as t
from collections import OrderedDict
from enum import Enum

from cratedb_toolkit.model import DatabaseAddress, TableAddress

logger = logging.getLogger(__name__)


class RetentionStrategy(Enum):
    """
    Enumerate list of retention strategies.
    """

    DELETE = "DELETE"
    REALLOCATE = "REALLOCATE"
    SNAPSHOT = "SNAPSHOT"

    def to_database(self) -> str:
        """
        In the database, strategy values are stored in lower case.
        """
        return str(self.value).lower()


@dataclasses.dataclass
class RetentionPolicy:
    """
    Manage the database representation of a "retention policy" entity.

    This layout has to be synchronized with the corresponding table definition
    per SQL DDL statement within `schema.sql`.
    """

    strategy: RetentionStrategy = dataclasses.field(
        metadata={"help": "Which kind of retention strategy to use or apply"},
    )
    tags: t.Optional[t.Union[list, set]] = dataclasses.field(
        default_factory=set,
        metadata={"help": "Tags for retention policy, used for grouping and filtering"},
    )
    table_schema: t.Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "The source table schema"},
    )
    table_name: t.Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "The source table name"},
    )
    partition_column: t.Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "The source table column name used for partitioning"},
    )
    retention_period: t.Optional[int] = dataclasses.field(
        default=None,
        metadata={
            "help": "Retention period in days. The number of days data gets "
            "retained before applying the retention policy."
        },
    )
    reallocation_attribute_name: t.Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "Name of the node-specific custom attribute, when targeting specific nodes"},
    )
    reallocation_attribute_value: t.Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "Value of the node-specific custom attribute, when targeting specific nodes"},
    )
    target_repository_name: t.Optional[str] = dataclasses.field(
        default=None,
        metadata={"help": "The name of a repository created with `CREATE REPOSITORY ...`, when targeting a repository"},
    )

    id: t.Optional[str] = dataclasses.field(  # noqa: A003
        default=None,
        metadata={"help": "The retention policy identifier"},
    )

    @property
    def table_fullname(self) -> str:
        return f'"{self.table_schema}"."{self.table_name}"'

    @classmethod
    def from_record(cls, record) -> "RetentionPolicy":
        strategy = RetentionStrategy(record["strategy"].upper())
        del record["strategy"]
        return cls(strategy=strategy, **record)

    def to_storage_dict(self, identifier: t.Optional[str] = None) -> t.Dict[str, str]:
        """
        Return representation suitable for storing into database table using SQLAlchemy.
        """

        # Serialize to dictionary.
        data = dataclasses.asdict(self)

        # Marshal strategy type.
        data["strategy"] = self.strategy.to_database()

        # Marshal list of tags to `OBJECT(DYNAMIC)` representation.
        tags = OrderedDict()
        for tag in data["tags"]:
            tags[tag] = "true"
        data["tags"] = tags

        # Optionally add identifier.
        if identifier is not None:
            data["id"] = identifier

        return data


@dataclasses.dataclass
class RetentionTask:
    """
    Represent a retention task at runtime.

    Specialized retention tasks, like `ReallocateRetentionTask`, derive from this class.

    It mostly contains attributes from `RetentionPolicy`, but a) offers marshalled
    values only, and b) omits some settings/parameters which have already been resolved
    by previous processing layers.
    """

    table_schema: str
    table_name: str
    table_fullname: str
    partition_column: str
    partition_value: str
    reallocation_attribute_name: str
    reallocation_attribute_value: str
    target_repository_name: str

    @classmethod
    def factory(cls, **kwargs):
        """
        Create an instance from keyword arguments.
        """
        return cls(**kwargs)

    @abc.abstractmethod
    def to_sql(self) -> t.Union[str, t.List[str]]:
        """
        Return one or more SQL statements, which resemble this task.
        """
        pass


def default_table_address():
    """
    The default address of the retention policy table.
    """
    schema = os.environ.get("CRATEDB_EXT_SCHEMA", "ext")
    return TableAddress(schema=schema, table="retention_policy")


@dataclasses.dataclass
class JobSettings:
    """
    Bundle all configuration and runtime settings.
    """

    # Database connection URI.
    database: DatabaseAddress = dataclasses.field(
        default_factory=lambda: DatabaseAddress.from_string("crate://localhost/")
    )

    # Retention strategy.
    strategy: t.Optional[RetentionStrategy] = None

    # Tags: For grouping, multi-tenancy, and more.
    tags: t.Set[str] = dataclasses.field(default_factory=set)

    # Retention cutoff timestamp.
    cutoff_day: t.Optional[str] = None

    # Where the retention policy table is stored.
    policy_table: TableAddress = dataclasses.field(default_factory=default_table_address)

    # Only pretend to invoke retention tasks.
    dry_run: t.Optional[bool] = False

    def to_dict(self):
        data = dataclasses.asdict(self)
        data["policy_table"] = self.policy_table
        return data



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


"""
Export the documents from a MongoDB collection as JSON, to be ingested into CrateDB.
"""

import calendar
import re
import sys
import typing as t
from datetime import datetime, timedelta

import bsonjs
import orjson as json
import pymongo.collection

_TZINFO_RE = re.compile(r"([+\-])?(\d\d):?(\d\d)")


def date_converter(value):
    if isinstance(value, int):
        return value
    dt = datetime.strptime(value[:-5], "%Y-%m-%dT%H:%M:%S.%f")
    iso_match = _TZINFO_RE.match(value[-5:])
    if iso_match:
        sign, hours, minutes = iso_match.groups()
        tzoffset = int(hours) * 3600 + int(minutes) * 60
        if sign == "-":
            dt = dt + timedelta(seconds=tzoffset)
        else:
            dt = dt - timedelta(seconds=tzoffset)
    else:
        raise Exception("Can't parse datetime string {0}".format(value))
    return calendar.timegm(dt.utctimetuple()) * 1000


def timestamp_converter(value):
    if len(str(value)) <= 10:
        return value * 1000
    return value


type_converter = {
    "date": date_converter,
    "timestamp": timestamp_converter,
    "undefined": lambda x: None,
}


def extract_value(value, parent_type=None):
    if isinstance(value, dict):
        if len(value) == 1:
            for k, v in value.items():
                if k.startswith("$"):
                    return extract_value(v, k.lstrip("$"))
        return {k.lstrip("$"): extract_value(v, parent_type) for (k, v) in value.items()}
    if isinstance(value, list):
        return [extract_value(v, parent_type) for v in value]
    if parent_type:
        converter = type_converter.get(parent_type)
        if converter:
            return converter(value)
    return value


def convert(d):
    newdict = {}
    del d["_id"]
    for k, v in d.items():
        newdict[k] = extract_value(v)
    return newdict


def collection_to_json(collection: pymongo.collection.Collection, file: t.IO[t.Any] = None):
    """
    Export a MongoDB collection's documents to standard JSON.
    The output is suitable to be consumed by the `cr8` program.

    collection
      a Pymongo collection object.

    file
      a file-like object (stream); defaults to the current sys.stdout.
    """
    file = file or sys.stdout.buffer
    for document in collection.find():
        bson_json = bsonjs.dumps(document.raw)
        json_object = json.loads(bson_json)
        file.write(json.dumps(convert(json_object)))
        file.write(b"\n")



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

"""
Export a schema definition from a MongoDB collection.

This will iterate over a collection (either totally, or partially) and build
up a description of the schema of the MongoDB collection.

Within the schema, each field in the collection will be described with two
fields:

- "count", being the number of entries in the collection that have this field.
- "types", being the types present for those entries.

For each type in a field's types, it will have a count that signifies the number
of entries of that field with that data type. If it is an object, it will also
contain a schema of the object's types. If it is an array, it will contain
a list of types that are present in the arrays, as well as their counts.

An example schema may look like:

{
    "count": 10,
    "document": {
        "ts": {
            "count": 10,
            "types": {"DATETIME": {"count": 10}
            }
        },
        "payload": {
            "count": 10,
            "types": {
                "OBJECT": {
                    "count": 10,
                    "document": {
                        "temp": {
                            "count": 10,
                            "types": {"FLOAT": {"count": 4}, "INTEGER": {"count": 6}}
                        }
                    }
                }
            }
        }
    }
}
"""

import typing as t

import bson
from pymongo.collection import Collection
from rich import progress

progressbar = progress.Progress(
    progress.TextColumn("{task.description} ", justify="left"),
    progress.BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}% ({task.completed}/{task.total})",
    "•",
    progress.TimeRemainingColumn(),
)


def extract_schema_from_collection(collection: Collection, partial: bool) -> t.Dict[str, t.Any]:
    """
    Extract a schema definition from a collection.

    If the extraction is partial, only the first document in the collection is
    used to create the schema.
    """

    schema: dict = {"count": 0, "document": {}}
    if partial:
        count = 1
    else:
        count = collection.estimated_document_count()
    with progressbar:
        t = progressbar.add_task(collection.name, total=count)
        try:
            for document in collection.find():
                schema["count"] += 1
                schema["document"] = extract_schema_from_document(document, schema["document"])
                progressbar.update(t, advance=1)
                if partial:
                    break
        except KeyboardInterrupt:
            return schema
    return schema


def extract_schema_from_document(document: dict, schema: dict):
    """
    Extract and update schema definition from a given document.
    """

    for k, v in document.items():
        if k not in schema:
            schema[k] = {"count": 0, "types": {}}

        item_type = get_type(v)
        if item_type not in schema[k]["types"]:
            if item_type == "OBJECT":
                schema[k]["types"][item_type] = {"count": 0, "document": {}}
            elif item_type == "ARRAY":
                schema[k]["types"][item_type] = {"count": 0, "types": {}}
            else:
                schema[k]["types"][item_type] = {"count": 0}

        schema[k]["count"] += 1
        schema[k]["types"][item_type]["count"] += 1
        if item_type == "OBJECT":
            schema[k]["types"][item_type]["document"] = extract_schema_from_document(
                v, schema[k]["types"][item_type]["document"]
            )
        elif item_type == "ARRAY":
            schema[k]["types"][item_type]["types"] = extract_schema_from_array(
                v, schema[k]["types"][item_type]["types"]
            )
    return schema


def extract_schema_from_array(array: list, schema: dict):
    """
    Extract and update a schema definition for a list.
    """

    for item in array:
        t = get_type(item)
        if t not in schema:
            if t == "OBJECT":
                schema[t] = {"count": 0, "document": {}}
            elif t == "ARRAY":
                schema[t] = {"count": 0, "types": {}}
            else:
                schema[t] = {"count": 0}

        schema[t]["count"] += 1
        if t == "OBJECT":
            schema[t]["document"] = extract_schema_from_document(item, schema[t]["document"])
        elif t == "ARRAY":
            schema[t]["types"] = extract_schema_from_array(item, schema[t]["types"])
    return schema


TYPES_MAP = {
    # bson types
    bson.ObjectId: "OID",
    bson.datetime.datetime: "DATETIME",
    bson.Timestamp: "TIMESTAMP",
    bson.int64.Int64: "INT64",
    # primitive types
    str: "STRING",
    bool: "BOOLEAN",
    int: "INTEGER",
    float: "FLOAT",
    # collection types
    list: "ARRAY",
    dict: "OBJECT",
}


def get_type(o):
    return TYPES_MAP.get(type(o), "UNKNOWN")



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
"""
Implements a retention policy by snapshotting expired partitions to a repository

A detailed tutorial is available at https://community.crate.io/t/cratedb-and-apache-airflow-building-a-data-retention-policy-using-external-snapshot-repositories/1001

Prerequisites
-------------
In CrateDB, tables for storing retention policies need to be created once manually.
See the file setup/schema.sql in this repository.
"""
import dataclasses
import logging

from cratedb_toolkit.retention.model import RetentionTask

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SnapshotRetentionTask(RetentionTask):
    """
    Represent a data retention task, using the `snapshot` strategy.
    """

    def to_sql(self):
        """
        Render as SQL statement.
        """
        # FIXME: S608 Possible SQL injection vector through string-based query construction
        sql = f"""
        CREATE SNAPSHOT "{self.target_repository_name}"."{self.table_schema}.{self.table_name}-{self.partition_value}"
        TABLE {self.table_fullname} PARTITION ({self.partition_column} = {self.partition_value})
        WITH ("wait_for_completion" = true);
        """  # noqa: S608
        sql2 = f"""
        DELETE FROM {self.table_fullname} WHERE {self.partition_column} = {self.partition_value};
        """  # noqa: S608
        return [sql, sql2]



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import datetime
import logging
import typing as t

import sqlalchemy as sa

from cratedb_toolkit.retention.model import JobSettings, RetentionPolicy, RetentionStrategy, RetentionTask
from cratedb_toolkit.retention.store import RetentionPolicyStore
from cratedb_toolkit.retention.strategy.delete import DeleteRetentionTask
from cratedb_toolkit.retention.strategy.reallocate import ReallocateRetentionJob, ReallocateRetentionTask
from cratedb_toolkit.retention.strategy.snapshot import SnapshotRetentionTask
from cratedb_toolkit.util.database import run_sql

logger = logging.getLogger(__name__)


class RetentionJob:
    """
    The main application, implementing a retention and expiration
    management subsystem for CrateDB.

    The retention job evaluates its configuration and runtime settings,
    and dispatches to corresponding retention strategy implementations.
    """

    def __init__(self, settings: JobSettings):
        # When no cutoff date is obtained from the caller, use `today()` as default value.
        if settings.cutoff_day is None:
            today = datetime.date.today().isoformat()
            logger.info(f"No cutoff date selected, will use today(): {today}")
            settings.cutoff_day = today

        # Runtime context settings.
        self.settings = settings

        # Retention policy store API.
        self.store = RetentionPolicyStore(settings=self.settings)

    def start(self):
        """
        Resolve retention policies to tasks, and invoke them.
        """
        logger.info(
            f"Connecting to database {self.settings.database.safe}, " f"table {self.settings.policy_table.fullname}"
        )

        msg = f"Starting data retention using '{self.settings.strategy}', cutoff day '{self.settings.cutoff_day}'"
        if self.settings.tags:
            msg += f", and tags '{self.settings.tags}'"
        logger.info(msg)

        # Invoke job.
        # TODO: Add audit logging.
        # TODO: Add job tracking.

        for task in self.get_retention_tasks():
            logger.info(f"Executing data retention task: {task}")

            sql_bunch: t.Iterable = task.to_sql()
            if not isinstance(sql_bunch, t.List):
                sql_bunch = [sql_bunch]

            # Run a sequence of SQL statements for this task.
            # Stop the sequence when a step fails.
            for sql in sql_bunch:
                if self.settings.dry_run:
                    logger.info(f"Pretending to execute SQL statement:\n{sql}")
                    continue
                try:
                    run_sql(dburi=self.settings.database.dburi, sql=sql)
                except Exception:
                    logger.exception(f"Data retention SQL statement failed: {sql}")
                    break

    def get_retention_tasks(self) -> t.Generator[RetentionTask, None, None]:
        """
        Derive retention tasks from policies. There will be one task per resolved partition.
        """

        # Sanity checks.
        if self.settings.strategy is None:
            raise ValueError("Unable to load retention policies without strategy")

        if self.settings.tags and not self.store.tags_exist(self.settings.tags):
            logger.warning(f"No retention policies found with tags: {sorted(self.settings.tags)}")
            return

        # Load policies, filtering by strategy and tags.
        policies = list(self.store.retrieve_policies(strategy=self.settings.strategy, tags=self.settings.tags))
        if not policies:
            logger.warning("No matching retention policies found")
            return

        for policy in policies:
            logger.info(f"Processing data retention policy: {policy}")

            # Verify data table exists.
            if not self.store.database.table_exists(policy.table_fullname):
                logger.warning(f"Data table not found: {policy.table_fullname}")
                continue

            # Render SQL statement to gather tasks.
            sql_renderer = TaskSqlRenderer(settings=self.settings, store=self.store, policy=policy)
            selectable = sql_renderer.render()
            results = self.store.query(selectable)
            if not results:
                logger.warning(f"No data to be retired from data table: {policy.table_fullname}")
                continue

            # Resolve task runner implementation.
            task_class: t.Type[RetentionTask]
            if policy.strategy is RetentionStrategy.DELETE:
                task_class = DeleteRetentionTask
            elif policy.strategy is RetentionStrategy.REALLOCATE:
                task_class = ReallocateRetentionTask
            elif policy.strategy is RetentionStrategy.SNAPSHOT:
                task_class = SnapshotRetentionTask
            else:
                raise NotImplementedError(f"Retention strategy {policy.strategy_type} not implemented yet")

            # Iterate tasks, and invoke task runner.
            for result in results:
                task = task_class.factory(
                    table_schema=policy.table_schema,
                    table_name=policy.table_name,
                    table_fullname=policy.table_fullname,
                    partition_column=policy.partition_column,
                    partition_value=result["partition_value"],
                    reallocation_attribute_name=policy.reallocation_attribute_name,
                    reallocation_attribute_value=policy.reallocation_attribute_value,
                    target_repository_name=policy.target_repository_name,
                )
                yield task


class TaskSqlRenderer:
    """
    Render SQL statement to gather retention tasks.
    """

    def __init__(self, settings: JobSettings, store: RetentionPolicyStore, policy: RetentionPolicy):
        """
        TODO: Get rid of `self.settings`?
        """
        self.settings = settings
        self.store = store
        self.policy = policy

    def render(self):
        """
        Render SQL statement.

        SQL DQL clause for selecting records from the retention policy database table,
        used for all strategy implementations. It can be interpolated into other SQL
        templates by using the `policy_dql` template variable.

        The retention policy database table is called `"ext"."retention_policy"` by default.
        """

        # Sanity checks.
        if self.settings.cutoff_day is None:
            raise ValueError("Unable to operate without cutoff date")

        # Formulate SQL statement to gather relevant partitions to be expired.
        # The `ORDER BY` is used to process partitions in chronological order, to not produce
        # temporary gaps. Imagine a table is partitioned by month: If partitions which are
        # selected for expiry are not processed in order, the recipe could first drop February,
        # then March, and then January, which would produce inconsistent gaps in the data, if
        # it was queried at the same time concurrently, while the expiration process is taking
        # place.
        policy = self.policy
        selectable = sa.text(
            f"""
        SELECT
            *,
            TRY_CAST(p.values['{policy.partition_column}'] AS BIGINT) AS partition_value
        FROM
            "information_schema"."table_partitions" AS p
        WHERE 1
            AND p.table_schema = '{policy.table_schema}'
            AND p.table_name = '{policy.table_name}'
            AND p.values['{policy.partition_column}'] <
                '{self.settings.cutoff_day}'::TIMESTAMP - '{policy.retention_period} days'::INTERVAL
        ORDER BY
            partition_value ASC
        """
        )
        return self.specialize(selectable)

    def specialize(self, selectable):
        """
        Specialize SQL statement.

        For example, the `REALLOCATE` strategy needs to wrap the foundational SQL statement
        into another one, in order to correlate the results with `sys.shards` and `sys.nodes`
        tables.
        """
        if self.settings.strategy is RetentionStrategy.REALLOCATE:
            # Acquire SQL template.
            sql_wrapper = ReallocateRetentionJob.SQL

            # Render SQL statement using classic string templating.
            tplvars = self.policy.to_storage_dict()
            sql = sql_wrapper.format(policy_dql=selectable, **tplvars)
            selectable = sa.text(sql)

        return selectable



import logging
import time
import typing as t
from pathlib import Path

from cratedb_toolkit.job.croud import jobs_list
from cratedb_toolkit.model import InputOutputResource, TableAddress
from cratedb_toolkit.util.croud import CroudCall, CroudWrapper

logger = logging.getLogger(__name__)


class CloudIoSpecs:
    """
    Define capabilities of CrateDB Cloud Import.
    """

    allowed_compressions = ["gzip", None]
    allowed_formats = ["csv", "json", "parquet"]


class CloudIo:
    """
    Wrap access to CrateDB Cloud Import API.
    """

    def __init__(self, cluster_id: str):
        self.cluster_id = cluster_id

    def load_resource(self, resource: InputOutputResource, target: TableAddress) -> t.Tuple[t.Dict, bool]:
        """
        Load resource from URL into CrateDB, using CrateDB Cloud infrastructure.
        """

        # Use `schema` and `table` when given, otherwise derive from input URL.
        if target.table is None:
            target.table = Path(resource.url).with_suffix("").stem

        logger.info(f"Loading data. source={resource}, target={target}")

        import_job = self.create_import_job(resource=resource, target=target)
        job_id = import_job["id"]
        time.sleep(0.15)
        outcome, success, found = self.find_job(job_id=job_id)
        if not found:
            logger.error(f"Job not found: {job_id}")
        if not outcome:
            outcome = import_job
        return outcome, success

    def find_job(self, job_id: str) -> t.Tuple[t.Dict, bool, bool]:
        """
        Find CrateDB Cloud job by identifier.
        """
        jobs = jobs_list(self.cluster_id)

        found = False
        success = False
        job_info: t.Dict = {}
        for job in jobs:
            if job["id"] == job_id:
                found = True
                job_info = job
                status = job["status"]
                message = job["progress"]["message"]
                message = f"{message} (status: {status})"
                if status == "SUCCEEDED":
                    success = True
                    logger.info(message)
                else:
                    logger.error(message)
                break

        fix_job_info_table_name(job_info)

        return job_info, success, found

    def create_import_job(self, resource: InputOutputResource, target: TableAddress) -> t.Dict[str, t.Any]:
        """
        Create CrateDB Cloud import job, using resource on filesystem or URL.

        croud clusters import-jobs create from-url --cluster-id e1e38d92-a650-48f1-8a70-8133f2d5c400 \
            --file-format csv --table my_table_name --url https://s3.amazonaws.com/my.import.data.gz --compression gzip
        """
        from croud.__main__ import import_job_create_common_args
        from croud.clusters.commands import import_jobs_create_from_file, import_jobs_create_from_url
        from croud.parser import Argument

        specs: t.List[Argument] = import_job_create_common_args
        arguments = []

        url_argument = Argument("--url", type=str, required=True, help="The URL the import file will be read from.")

        file_id_argument = Argument(
            "--file-id",
            type=str,
            required=False,
            help="The file ID that will be used for the "
            "import. If not specified then --file-path"
            " must be specified. "
            "Please refer to `croud organizations "
            "files` for more info.",
        )
        file_path_argument = Argument(
            "--file-path",
            type=str,
            required=False,
            help="The file in your local filesystem that "
            "will be used. If not specified then "
            "--file-id must be specified. "
            "Please note the file will become visible "
            "under `croud organizations files list`.",
        )

        # Compute command-line arguments for invoking `croud`.
        path = Path(resource.url)

        # Honor `schema` argument.
        table = target.table
        if target.schema is not None:
            table = f'"{target.schema}"."{target.table}"'

        if resource.compression is None:
            if ".gz" in path.suffixes or ".gzip" in path.suffixes:
                resource.compression = "gzip"

        if resource.format is None:
            if ".csv" in path.suffixes or ".tsv" in path.suffixes:
                resource.format = "csv"
            elif ".json" in path.suffixes or ".jsonl" in path.suffixes or ".ndjson" in path.suffixes:
                resource.format = "json"
            elif ".parquet" in path.suffixes or ".pq" in path.suffixes:
                resource.format = "parquet"

        # Sanity checks.
        if resource.compression not in CloudIoSpecs.allowed_compressions:
            raise NotImplementedError(
                f"Unknown input file format compression: {resource.compression}. "
                f"Use one of: {CloudIoSpecs.allowed_compressions}"
            )
        if resource.format not in CloudIoSpecs.allowed_formats:
            raise NotImplementedError(
                f"Unknown input file format suffixes: {path.suffixes} ({resource.format}). "
                f"Use one of: {CloudIoSpecs.allowed_formats}"
            )

        # TODO: Sanitize table name. Which characters are allowed?
        if path.exists():
            specs.append(file_path_argument)
            specs.append(file_id_argument)
            arguments = [
                f"--cluster-id={self.cluster_id}",
                f"--file-path={resource.url}",
                f"--table={table}",
                f"--file-format={resource.format}",
            ]
            fun = import_jobs_create_from_file
        else:
            specs.append(url_argument)
            arguments = [
                f"--cluster-id={self.cluster_id}",
                f"--url={resource.url}",
                f"--table={table}",
                f"--file-format={resource.format}",
            ]
            fun = import_jobs_create_from_url

        if resource.compression is not None:
            arguments += [f"--compression={resource.compression}"]

        call = CroudCall(
            fun=fun,
            specs=specs,
            arguments=arguments,
        )

        wr = CroudWrapper(call=call)
        job_info = wr.invoke()

        # Adjust full-qualified table name by adding appropriate quotes.
        # FIXME: Remove after flaw has been fixed.
        fix_job_info_table_name(job_info)

        return job_info


def fix_job_info_table_name(job_info: t.Dict[str, t.Any]):
    # FIXME: Remove after upstream has fixed the flaw.
    #        Currently, the API returns `testdrive.pems-1`, but that can not be used at all.
    #        So, converge it into `"testdrive"."pems-1"`.
    if "destination" in job_info and "table" in job_info["destination"]:
        table = job_info["destination"]["table"]
        if '"' not in table and "." in table:
            schema, table = table.split(".")
            table = f'"{schema}"."{table}"'
            job_info["destination"]["table"] = table



from typing import Generator

import pytest

try:
    from cratedb_toolkit.testing.testcontainers.cratedb import CrateDBTestAdapter

    @pytest.fixture(scope="session")
    def cratedb_service() -> Generator[CrateDBTestAdapter, None, None]:
        """
        Provide a CrateDB service instance to the test suite.
        """
        db = CrateDBTestAdapter()
        db.start()
        yield db
        db.stop()

except ModuleNotFoundError as ex:
    message = str(ex)

    @pytest.fixture(scope="session")
    def cratedb_service():
        """
        An error handling surrogate used when dependencies are not satisfied.
        In order to use it, invoke `pip install 'cratedb-toolkit[testing]'`.
        """
        raise NotImplementedError(message)



from cratedb_toolkit.util.croud import CroudCall, CroudWrapper


def jobs_list(cratedb_cloud_cluster_id: str, output_format: str = None, decode_output: bool = True):
    from croud.clusters.commands import import_jobs_list
    from croud.parser import Argument

    call = CroudCall(
        fun=import_jobs_list,
        specs=[Argument("--cluster-id", type=str, required=True, help="The cluster the import jobs belong to.")],
        arguments=[
            f"--cluster-id={cratedb_cloud_cluster_id}",
        ],
    )

    wr = CroudWrapper(call, output_format=output_format, decode_output=decode_output)
    return wr.invoke()



import typing as t
from pathlib import Path

from cr8.insert_json import insert_json


def cr8_insert_json(infile: t.Union[str, Path, t.IO[t.Any]], hosts: str, table: str):
    return insert_json(table=table, bulk_size=5_000, hosts=hosts, infile=infile, output_fmt="json")



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import logging
import os
from typing import Any

from testcontainers.core.container import DockerContainer

logger = logging.getLogger(__name__)


# from sqlalchemy.util.langhelpers
# from paste.deploy.converters
def asbool(obj: Any) -> bool:
    if isinstance(obj, str):
        obj = obj.strip().lower()
        if obj in ["true", "yes", "on", "y", "t", "1"]:
            return True
        elif obj in ["false", "no", "off", "n", "f", "0"]:
            return False
        else:
            raise ValueError("String is not true/false: %r" % obj)
    return bool(obj)


class ExtendedDockerContainer(DockerContainer):
    """
    An extended Testcontainer.

    - Provide convenience methods for getting the Docker-internal endpoint address.
      TODO: Maybe rename to `get_bridge_host_*`?
    """

    def get_real_host_ip(self) -> str:
        """
        To let containers talk to each other, explicitly provide the real IP address
        of the container. In corresponding jargon, it appears to be the "bridge IP".
        """
        return self.get_docker_client().bridge_ip(self._container.id)

    def get_real_host_address(self) -> str:
        """
        Provide Docker-internal full endpoint address `<host>:<port>` of the service.
        For example, `172.17.0.4:9000`.
        """
        return f"{self.get_real_host_ip()}:{self.port_to_expose}"


class KeepaliveContainer(DockerContainer):
    """
    Improved `start()`/`stop()` methods, supporting service-keepalive.

    In order to keep the service running where it normally would be torn down,
    define the `TC_KEEPALIVE` environment variable.
    """

    KEEPALIVE = asbool(os.environ.get("TC_KEEPALIVE", False))

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        self.keepalive = self.KEEPALIVE
        if "keepalive" in kwargs:
            self.keepalive = kwargs["keepalive"]
            del kwargs["keepalive"]
        super().__init__(*args, **kwargs)

    def start(self):
        """
        Improved `start()` method, supporting service-keepalive.

        In order to keep the service running where it normally would be torn down,
        define the `CRATEDB_KEEPALIVE` or `TC_KEEPALIVE` environment variables.
        """

        self._configure()

        if self._name is None:
            raise ValueError(
                "KeepaliveContainer does not support unnamed containers. Use `.with_name()` to assign a name."
            )

        docker_client = self.get_docker_client()

        # Check if container is already running, and whether it should be reused.
        logger.info(f"Searching for container: {self._name}")
        containers = docker_client.client.api.containers(all=True, filters={"name": self._name})

        if not containers:
            logger.info(f"Pulling image: {self.image}")
            docker_client.client.images.pull(self.image)
            logger.info(f"Creating container from image: {self.image}")
            self._container = docker_client.run(
                self.image,
                command=self._command,
                detach=True,
                environment=self.env,
                ports=self.ports,
                name=self._name,
                volumes=self.volumes,
                **self._kwargs,
            )
            logger.info(f"Container created: {self._container.name}")
        else:
            container_id = containers[0]["Id"]
            container_names = containers[0]["Names"]
            logger.info(f"Found container for reuse: {container_id} ({container_names})")
            self._container = docker_client.client.containers.get(container_id)
            container_name = self._container.name
            if self._container.status != "running":
                logger.info(f"Starting container: {container_id} ({container_name})")
                self._container.start()

        self._connect()
        return self

    def stop(self, **kwargs):
        """
        Shut down container again, unless "keepalive" is enabled.
        """
        if not self.keepalive:
            logger.info("Stopping container")
            return super().stop()
        return None



import argparse
import json

import rich

from cratedb_toolkit import __version__
from cratedb_toolkit.io.mongodb.core import export, extract, translate


def extract_parser(subargs):
    parser = subargs.add_parser("extract", help="Extract a schema from a MongoDB database")
    parser.add_argument("--url", default="mongodb://localhost:27017", help="MongoDB URL")
    parser.add_argument("--host", default="localhost", help="MongoDB host")
    parser.add_argument("--port", default=27017, help="MongoDB port")
    parser.add_argument("--database", required=True, help="MongoDB database")
    parser.add_argument("--collection", help="MongoDB collection to create a schema for")
    parser.add_argument(
        "--scan",
        choices=["full", "partial"],
        help="Whether to fully scan the MongoDB collections or only partially.",
    )
    parser.add_argument("-o", "--out", default="mongodb_schema.json")


def translate_parser(subargs):
    parser = subargs.add_parser(
        "translate",
        help="Translate a MongoDB schema definition to a CrateDB table schema",
    )
    parser.add_argument("-i", "--infile", help="The JSON file to read the MongoDB schema from")


def export_parser(subargs):
    parser = subargs.add_parser("export", help="Export a MongoDB collection as plain JSON")
    parser.add_argument("--url", default="mongodb://localhost:27017", help="MongoDB URL")
    parser.add_argument("--collection", required=True)
    parser.add_argument("--host", default="localhost", help="MongoDB host")
    parser.add_argument("--port", default=27017, help="MongoDB port")
    parser.add_argument("--database", required=True, help="MongoDB database")


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="print package version of pyproject_fmt",
        version=f"%(prog)s ({__version__})",
    )
    subparsers = parser.add_subparsers(dest="command")
    extract_parser(subparsers)
    translate_parser(subparsers)
    export_parser(subparsers)
    return parser.parse_args()


def extract_to_file(args):
    """
    Extract a schema or set of schemas from MongoDB collections into a JSON file.
    """

    schema = extract(args)
    rich.print(f"\nWriting resulting schema to {args.out}...")
    with open(args.out, "w") as out:
        json.dump(schema, out, indent=4)
    rich.print("[green bold]Done![/green bold]")


def translate_from_file(args):
    """
    Read in a JSON file and extract the schema from it.
    """

    with open(args.infile) as f:
        schema = json.load(f)
        translate(schema)


def export_to_stdout(args):
    export(args)


def main():
    rich.print("\n[green bold]MongoDB[/green bold] -> [blue bold]CrateDB[/blue bold] Exporter :: Schema Extractor\n\n")
    args = get_args()
    if args.command == "extract":
        extract_to_file(args)
    elif args.command == "translate":
        translate_from_file(args)
    elif args.command == "export":
        export_to_stdout(args)



import click

from cratedb_toolkit.cluster.util import get_cluster_info
from cratedb_toolkit.util.cli import boot_click
from cratedb_toolkit.util.crash import get_crash_output_formats, run_crash

output_formats = get_crash_output_formats()


@click.command()
@click.option(
    "--cluster-id", envvar="CRATEDB_CLOUD_CLUSTER_ID", type=str, required=True, help="CrateDB Cloud cluster identifier"
)
@click.option("--username", envvar="CRATEDB_USERNAME", type=str, required=False, help="Username for CrateDB cluster")
@click.option("--password", envvar="CRATEDB_PASSWORD", type=str, required=False, help="Password for CrateDB cluster")
@click.option(
    "--schema",
    envvar="CRATEDB_SCHEMA",
    type=str,
    required=False,
    help="Default schema for statements if schema is not explicitly stated within queries",
)
@click.option("--command", type=str, required=False, help="SQL command")
@click.option(
    "--format",
    "format_",
    type=click.Choice(output_formats, case_sensitive=False),
    required=False,
    help="The output format of the database response",
)
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.version_option()
@click.pass_context
def cli(
    ctx: click.Context,
    cluster_id: str,
    username: str,
    password: str,
    schema: str,
    command: str,
    format_: str,
    verbose: bool,
    debug: bool,
):
    """
    Start an interactive database shell, or invoke SQL commands.

    TODO: Only talks to CrateDB Cloud for now. Also implement for standalone CrateDB servers.
    TODO: Learn/forward more options of `crash`.
    """
    boot_click(ctx, verbose, debug)

    cluster_info = get_cluster_info(cluster_id=cluster_id)
    cratedb_http_url = cluster_info.cloud["url"]

    run_crash(
        hosts=cratedb_http_url,
        username=username,
        password=password,
        schema=schema,
        command=command,
        output_format=format_,
    )



import os
import zipfile
from unittest.mock import patch


def load_dataset_kaggle(dataset: str, path: str, file_name: str = None):
    """
    Download complete dataset or individual files from Kaggle.

    To make it work, please supply Kaggle authentication information.

    Synopsis:

        export KAGGLE_USERNAME=foobar
        export KAGGLE_KEY=2b1dac2af55caaf1f34df76236fada5f
        kaggle datasets list -s weather
    """
    from kaggle import api as kaggle_api

    if file_name is None:
        with patch("os.remove"):
            kaggle_api.dataset_download_files(dataset=dataset, path=path, unzip=True)
    else:
        kaggle_api.dataset_download_file(dataset=dataset, file_name=file_name, path=path)
        outfile = os.path.join(path, file_name + ".zip")
        with zipfile.ZipFile(outfile) as z:
            z.extractall(path)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

from testcontainers.mongodb import MongoDbContainer

from cratedb_toolkit.testing.testcontainers.util import KeepaliveContainer


class MongoDbContainerWithKeepalive(KeepaliveContainer, MongoDbContainer):
    """
    A Testcontainer for MongoDB with improved configurability.

    It honors the `TC_KEEPALIVE` and `MONGODB_VERSION` environment variables.

    Defining `TC_KEEPALIVE` will set a signal not to shut down the container
    after running the test cases, in order to speed up subsequent invocations.

    `MONGODB_VERSION` will define the designated MongoDB version, which is
    useful when used within a test matrix. Its default value is `latest`.
    """

    MONGODB_VERSION = os.environ.get("MONGODB_VERSION", "latest")

    def __init__(
        self,
        image: str = f"mongo:{MONGODB_VERSION}",
        **kwargs,
    ) -> None:
        super().__init__(image=image, **kwargs)
        self.with_name("testcontainers-mongodb")



import sys

import click

from cratedb_toolkit.job.croud import jobs_list
from cratedb_toolkit.util.croud import get_croud_output_formats

try:
    output_formats = get_croud_output_formats()
except ImportError:
    output_formats = ["UNKNOWN"]


@click.command(name="list-jobs")
@click.option(
    "--cluster-id", envvar="CRATEDB_CLOUD_CLUSTER_ID", type=str, required=True, help="CrateDB Cloud cluster identifier"
)
@click.option(
    "--format",
    "format_",
    type=click.Choice(output_formats, case_sensitive=False),
    required=False,
    help="The output format for the result of the operation",
)
@click.pass_context
def cli_list_jobs(ctx: click.Context, cluster_id: str, format_: str):
    """
    List CrateDB Cloud jobs.
    """
    print(jobs_list(cluster_id, output_format=format_, decode_output=False), file=sys.stdout)  # noqa: T201



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import typing as t

from testcontainers.minio import MinioContainer

from cratedb_toolkit.testing.testcontainers.util import ExtendedDockerContainer


class ExtendedMinioContainer(ExtendedDockerContainer, MinioContainer):
    """
    An extended Testcontainer for MinIO, emulating AWS S3.

    Features
    - Use the `latest` OCI image from https://quay.io/.
    - Provide convenience methods for getting the Docker-internal endpoint address.

    References
    - https://en.wikipedia.org/wiki/Object_storage
    - https://en.wikipedia.org/wiki/Amazon_S3
    - https://github.com/minio/minio
    """

    def __init__(self, *args, **kwargs):
        # Use most recent stable release of MinIO.
        image = "quay.io/minio/minio:latest"
        kwargs.setdefault("image", image)

        super().__init__(*args, **kwargs)

    def list_object_names(self, bucket_name: str) -> t.List[str]:
        """
        Return list of object names within given bucket.
        """
        objects = self.get_client().list_objects(bucket_name=bucket_name)
        return [obj.object_name for obj in objects]



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

from influxdb_client import InfluxDBClient
from testcontainers.core.config import MAX_TRIES
from testcontainers.core.generic import DbContainer
from testcontainers.core.waiting_utils import wait_container_is_ready, wait_for_logs

from cratedb_toolkit.testing.testcontainers.util import KeepaliveContainer


class InfluxDB2Container(KeepaliveContainer, DbContainer):
    """
    InfluxDB database container.

    - https://en.wikipedia.org/wiki/Influxdb

    Example:

        The example spins up an InfluxDB2 database instance.
    """

    INFLUXDB_VERSION = os.environ.get("INFLUXDB_VERSION", "latest")

    ORGANIZATION = "example"
    TOKEN = "token"  # noqa: S105

    # TODO: Dual-port use with 8083+8086.
    def __init__(
        self,
        image: str = f"influxdb:{INFLUXDB_VERSION}",
        port: int = 8086,
        dialect: str = "influxdb2",
        **kwargs,
    ) -> None:
        super().__init__(image=image, **kwargs)

        self._name = "testcontainers-influxdb"  # -{os.getpid()}

        self.port_to_expose = port
        self.dialect = dialect

        self.with_exposed_ports(self.port_to_expose, 8083)

        self.debug = False

    def _configure(self) -> None:
        self.with_env("DOCKER_INFLUXDB_INIT_MODE", "setup")
        self.with_env("DOCKER_INFLUXDB_INIT_USERNAME", "admin")
        self.with_env("DOCKER_INFLUXDB_INIT_PASSWORD", "secret1234")
        self.with_env("DOCKER_INFLUXDB_INIT_ORG", self.ORGANIZATION)
        self.with_env("DOCKER_INFLUXDB_INIT_BUCKET", "default")
        self.with_env("DOCKER_INFLUXDB_INIT_ADMIN_TOKEN", self.TOKEN)

    def get_connection_url(self, host=None) -> str:
        return super()._create_connection_url(
            dialect="http",
            username=self.ORGANIZATION,
            password=self.TOKEN,
            host=host,
            port=self.port_to_expose,
        )

    @wait_container_is_ready()
    def _connect(self) -> InfluxDBClient:
        # TODO: Better use a network connectivity health check?
        #       In `testcontainers-java`, there is the `HttpWaitStrategy`.
        wait_for_logs(self, predicate="Listening.*tcp-listener.*8086", timeout=MAX_TRIES)
        return InfluxDBClient(url=self.get_connection_url(), org=self.ORGANIZATION, token=self.TOKEN, debug=self.debug)

    def get_connection_client(self) -> InfluxDBClient:
        return self._connect()



import dataclasses
import typing as t
from copy import deepcopy

from boltons.urlutils import URL

from cratedb_toolkit.util.database import decode_database_table


@dataclasses.dataclass
class DatabaseAddress:
    """
    Manage a database address, which is either a SQLAlchemy-
    compatible database URI, or a regular HTTP URL.
    """

    uri: URL

    @classmethod
    def from_string(cls, url):
        """
        Factory method to create an instance from an SQLAlchemy database URL in string format.
        """
        return cls(uri=URL(url))

    @classmethod
    def from_httpuri(cls, url):
        """
        Factory method to create an instance from an HTTP URL in string format.
        """
        uri = URL(url)
        if uri.scheme == "https":
            uri.query_params["ssl"] = "true"
        uri.scheme = "crate"
        return cls(uri=uri)

    @property
    def dburi(self) -> str:
        """
        Return a string representation of the database URI.
        """
        return str(self.uri)

    @property
    def httpuri(self) -> str:
        """
        Return the `http(s)://` variant of the database URI.
        """
        uri = deepcopy(self.uri)
        uri.scheme = "http"
        if "ssl" in uri.query_params:
            if uri.query_params["ssl"]:
                uri.scheme = "https"
            del uri.query_params["ssl"]
        return str(uri)

    @property
    def safe(self):
        """
        Return a string representation of the database URI, safe for printing.
        The password is stripped from the URL, and replaced by `REDACTED`.
        """
        uri = deepcopy(self.uri)
        uri.password = "REDACTED"  # noqa: S105
        return str(uri)

    def decode(self) -> t.Tuple[URL, "TableAddress"]:
        """
        Decode database and table names, and sanitize database URI.
        """
        database, table = decode_database_table(self.dburi)
        uri = deepcopy(self.uri)
        uri.path = ""
        return uri, TableAddress(database, table)


@dataclasses.dataclass
class TableAddress:
    """
    Manage a table address, which is made of "<schema>"."<table>".
    """

    schema: t.Optional[str] = None
    table: t.Optional[str] = None

    @property
    def fullname(self):
        if self.schema is None and self.table is None:
            raise ValueError("Uninitialized table address can not be serialized")
        if self.schema and self.table:
            return f'"{self.schema}"."{self.table}"'
        else:
            return f'"{self.table}"'


@dataclasses.dataclass
class ClusterInformation:
    """
    Manage a database cluster's information.
    """

    cratedb: t.Any = dataclasses.field(default_factory=dict)
    cloud: t.Dict[str, t.Any] = dataclasses.field(default_factory=dict)

    def asdict(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass
class InputOutputResource:
    """
    Manage information about an input or output resource.
    """

    url: str
    format: t.Optional[str] = None  # noqa: A003
    compression: t.Optional[str] = None



class GuidingTexts:
    """
    TODO: Add more richness / guidance to the text output.
    """

    def __init__(self, admin_url: str = None, table_name: str = None):
        self.admin_url = admin_url
        self.table_name = table_name

    def success(self):
        return f"""
        Excellent, that worked well.

        Now, you may want to inquire your data. To do that, use either CrateDB Admin UI,
        or connect on your terminal using `crash`, `ctk shell`, or `psql`.

        The CrateDB Admin UI for your cluster is available at [1]. To easily inspect a
        few samples of your imported data, or to check the cardinality of your database
        table, run [2] or [3]. If you want to export your data again, see [4].

        [1] {self.admin_url}
        [2] ctk shell --command 'SELECT * FROM {self.table_name} LIMIT 10;'
        [3] ctk shell --command 'SELECT COUNT(*) FROM {self.table_name};'
        [4] https://community.cratedb.com/t/cratedb-cloud-news-simple-data-export/1556
        """  # noqa: S608

    def error(self):
        return """
        That went south.

        If you can share your import source, we will love to hear from you on our community
        forum [1]. Otherwise, please send us an email [2] about the flaw you've discovered.
        To learn more about the data import feature, see [3].

        [1] https://community.cratedb.com/
        [2] support@crate.io
        [3] https://community.cratedb.com/t/importing-data-to-cratedb-cloud-clusters/1467
        """



# Copyright (c) 2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import io
import os
import typing as t
from pathlib import Path

import sqlalchemy as sa
import sqlparse
from boltons.urlutils import URL
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.sql.elements import AsBoolean

from cratedb_toolkit.util.data import str_contains

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore[assignment]


def run_sql(dburi: str, sql: str, records: bool = False):
    return DatabaseAdapter(dburi=dburi).run_sql(sql=sql, records=records)


class DatabaseAdapter:
    """
    Wrap SQLAlchemy connection to database.
    """

    def __init__(self, dburi: str):
        self.dburi = dburi
        # TODO: Make `echo=True` configurable.
        self.engine = sa.create_engine(self.dburi, echo=False)
        self.connection = self.engine.connect()

    def run_sql(self, sql: t.Union[str, Path, io.IOBase], records: bool = False, ignore: str = None):
        """
        Run SQL statement, and return results, optionally ignoring exceptions.
        """

        sql_effective: str
        if isinstance(sql, str):
            sql_effective = sql
        elif isinstance(sql, Path):
            sql_effective = sql.read_text()
        elif isinstance(sql, io.IOBase):
            sql_effective = sql.read()
        else:
            raise TypeError("SQL statement type must be either string, Path, or IO handle")

        try:
            return self.run_sql_real(sql=sql_effective, records=records)
        except Exception as ex:
            if not ignore:
                raise
            if ignore not in str(ex):
                raise

    def run_sql_real(self, sql: str, records: bool = False):
        """
        Invoke SQL statement, and return results.
        """
        results = []
        with self.engine.connect() as connection:
            for statement in sqlparse.split(sql):
                result = connection.execute(sa.text(statement))
                data: t.Any
                if records:
                    rows = result.mappings().fetchall()
                    data = [dict(row.items()) for row in rows]
                else:
                    data = result.fetchall()
                results.append(data)

        # Backward-compatibility.
        if len(results) == 1:
            return results[0]
        else:
            return results

    def count_records(self, tablename_full: str, errors: Literal["raise", "ignore"] = "raise"):
        """
        Return number of records in table.
        """
        sql = f"SELECT COUNT(*) AS count FROM {tablename_full};"  # noqa: S608
        try:
            results = self.run_sql(sql=sql)
        except ProgrammingError as ex:
            is_candidate = "RelationUnknown" not in str(ex)
            if is_candidate and errors == "raise":
                raise
            return 0
        return results[0][0]

    def table_exists(self, tablename_full: str) -> bool:
        """
        Check whether given table exists.
        """
        sql = f"SELECT 1 FROM {tablename_full} LIMIT 1;"  # noqa: S608
        try:
            self.run_sql(sql=sql)
            return True
        except Exception:
            return False

    def refresh_table(self, tablename_full: str):
        """
        Run a `REFRESH TABLE ...` command.
        """
        sql = f"REFRESH TABLE {tablename_full};"  # noqa: S608
        self.run_sql(sql=sql)
        return True

    def prune_table(self, tablename_full: str, errors: Literal["raise", "ignore"] = "raise"):
        """
        Run a `DELETE FROM ...` command.
        """
        sql = f"DELETE FROM {tablename_full};"  # noqa: S608
        try:
            self.run_sql(sql=sql)
        except ProgrammingError as ex:
            is_candidate = "RelationUnknown" not in str(ex)
            if is_candidate and errors == "raise":
                raise
            return False
        return True

    def drop_table(self, tablename_full: str):
        """
        Run a `DROP TABLE ...` command.
        """
        sql = f"DROP TABLE IF EXISTS {tablename_full};"  # noqa: S608
        self.run_sql(sql=sql)
        return True

    def drop_repository(self, name: str):
        """
        Drop snapshot repository.
        """
        # TODO: DROP REPOSITORY IF EXISTS
        try:
            sql = f"DROP REPOSITORY {name};"
            self.run_sql(sql)
        except ProgrammingError as ex:
            if not str_contains(ex, "RepositoryUnknownException", "RepositoryMissingException"):
                raise

    def ensure_repository_fs(
        self,
        name: str,
        typename: str,
        location: str,
        drop: bool = False,
    ):
        """
        Make sure the repository exists, and optionally drop it upfront.
        """
        if drop:
            self.drop_repository(name)

        # TODO: CREATE REPOSITORY IF NOT EXISTS
        sql = f"""
            CREATE REPOSITORY
                {name}
            TYPE
                {typename}
            WITH (
                location   = '{location}'
            );
        """
        self.run_sql(sql)

    def ensure_repository_s3(
        self,
        name: str,
        typename: str,
        protocol: str,
        endpoint: str,
        access_key: str,
        secret_key: str,
        bucket: str,
        drop: bool = False,
    ):
        """
        Make sure the repository exists, and optionally drop it upfront.
        """
        if drop:
            self.drop_repository(name)

        # TODO: CREATE REPOSITORY IF NOT EXISTS
        sql = f"""
            CREATE REPOSITORY
                {name}
            TYPE
                {typename}
            WITH (
                protocol   = '{protocol}',
                endpoint   = '{endpoint}',
                access_key = '{access_key}',
                secret_key = '{secret_key}',
                bucket     = '{bucket}'
            );
        """
        self.run_sql(sql)

    def ensure_repository_az(
        self,
        name: str,
        typename: str,
        protocol: str,
        endpoint: str,
        account: str,
        key: str,
        container: str,
        drop: bool = False,
    ):
        """
        Make sure the repository exists, and optionally drop it upfront.
        """
        if drop:
            self.drop_repository(name)

        # TODO: CREATE REPOSITORY IF NOT EXISTS
        sql = f"""
            CREATE REPOSITORY
                {name}
            TYPE
                {typename}
            WITH (
                protocol   = '{protocol}',
                endpoint   = '{endpoint}',
                account    = '{account}',
                key        = '{key}',
                container  = '{container}'
            );
        """
        self.run_sql(sql)

    def import_csv_pandas(
        self, filepath: t.Union[str, Path], tablename: str, index=False, chunksize=1000, if_exists="replace"
    ):
        """
        Import CSV data using pandas.
        """
        import pandas as pd
        from crate.client.sqlalchemy.support import insert_bulk

        df = pd.read_csv(filepath)
        with self.engine.connect() as connection:
            return df.to_sql(
                tablename, connection, index=index, chunksize=chunksize, if_exists=if_exists, method=insert_bulk
            )

    def import_csv_dask(
        self,
        filepath: t.Union[str, Path],
        tablename: str,
        index=False,
        chunksize=1000,
        if_exists="replace",
        npartitions: int = None,
        progress: bool = False,
    ):
        """
        Import CSV data using Dask.
        """
        import dask.dataframe as dd
        import pandas as pd
        from crate.client.sqlalchemy.support import insert_bulk

        # Set a few defaults.
        npartitions = npartitions or os.cpu_count()

        if progress:
            from dask.diagnostics import ProgressBar

            pbar = ProgressBar()
            pbar.register()

        # Load data into database.
        df = pd.read_csv(filepath)
        ddf = dd.from_pandas(df, npartitions=npartitions)
        return ddf.to_sql(
            tablename,
            uri=self.dburi,
            index=index,
            chunksize=chunksize,
            if_exists=if_exists,
            method=insert_bulk,
            parallel=True,
        )


def sa_is_empty(thing):
    """
    When a WHERE criteria clause is empty, i.e. it contains only an
    `and_` element, let's consider it to be empty.

    TODO: Verify this. How to actually compare SQLAlchemy elements by booleanness?
    """
    return isinstance(thing, AsBoolean)


def decode_database_table(url: str) -> t.Tuple[str, str]:
    """
    Decode database and table names from database URI path and/or query string.

    Variants:

        /<database>/<table>
        ?database=<database>&table=<table>

    TODO: Synchronize with `influxio.model.decode_database_table`.
          This one uses `boltons`, the other one uses `yarl`.
    """
    url_ = URL(url)
    try:
        database, table = url_.path.strip("/").split("/")
    except ValueError as ex:
        if "too many values to unpack" not in str(ex) and "not enough values to unpack" not in str(ex):
            raise
        database = url_.query_params.get("database")
        table = url_.query_params.get("table")
        if url_.scheme == "crate" and not database:
            database = url_.query_params.get("schema")
    return database, table



from cratedb_toolkit.model import ClusterInformation


def get_cluster_info(cluster_id: str) -> ClusterInformation:
    cluster_info = ClusterInformation()
    try:
        from cratedb_toolkit.cluster.croud import CloudCluster

        cc = CloudCluster(cluster_id=cluster_id)
        cluster_info.cloud = cc.get_info()
    except ImportError:
        pass
    return cluster_info



import click
from click_aliases import ClickAliasedGroup

from cratedb_toolkit.util.cli import boot_click

from .cluster.cli import cli as cloud_cli
from .io.cli import cli as io_cli
from .job.cli import cli_list_jobs
from .shell.cli import cli as shell_cli


@click.group(cls=ClickAliasedGroup)  # type: ignore[arg-type]
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    return boot_click(ctx, verbose, debug)


cli.add_command(cloud_cli, name="cluster")
cli.add_command(io_cli, name="load")
cli.add_command(shell_cli, name="shell")
cli.add_command(cli_list_jobs)



# Copyright (c) 2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import logging

import colorlog
from colorlog.escape_codes import escape_codes


def setup_logging(level=logging.INFO, verbose: bool = False):
    reset = escape_codes["reset"]
    log_format = f"%(asctime)-15s [%(name)-36s] %(log_color)s%(levelname)-8s:{reset} %(message)s"

    handler = colorlog.StreamHandler()
    handler.setFormatter(colorlog.ColoredFormatter(log_format))

    logging.basicConfig(format=log_format, level=level, handlers=[handler])

    # Enable SQLAlchemy logging.
    if verbose:
        logging.getLogger("sqlalchemy").setLevel(level)

    logging.getLogger("crate.client").setLevel(level)
    logging.getLogger("urllib3.connectionpool").setLevel(level)

    # logging.getLogger("docker.auth").setLevel(logging.INFO)  # noqa: ERA001

    # Tame Faker spamming the logs.
    # https://github.com/joke2k/faker/issues/753#issuecomment-491402018
    logging.getLogger("faker").setLevel(logging.ERROR)



import itertools

import sqlalchemy as sa
from sqlalchemy.event import listen


def polyfill_autoincrement():
    """
    Configure SQLAlchemy model columns with an alternative to `autoincrement=True`.

    In this case, use a random identifier: Nagamani19, a short, unique,
    non-sequential identifier based on Hashids.

    TODO: Submit patch to `crate-python`, to be enabled by a
          dialect parameter `crate_polyfill_autoincrement` or such.
    """
    import sqlalchemy.sql.schema as schema
    from sqlalchemy import func

    init_dist = schema.Column.__init__

    def __init__(self, *args, **kwargs):
        if "autoincrement" in kwargs:
            del kwargs["autoincrement"]
            if "default" not in kwargs:
                kwargs["default"] = func.now()
        init_dist(self, *args, **kwargs)

    schema.Column.__init__ = __init__  # type: ignore[method-assign]


def check_uniqueness_factory(sa_entity, *attribute_names):
    """
    Run a manual column value uniqueness check on a table, and raise an IntegrityError if applicable.

    CrateDB does not support the UNIQUE constraint on columns. This attempts to emulate it.

    TODO: Submit patch to `crate-python`, to be enabled by a
          dialect parameter `crate_polyfill_unique` or such.
    """

    # Synthesize a canonical "name" for the constraint,
    # composed of all column names involved.
    constraint_name: str = "-".join(attribute_names)

    def check_uniqueness(mapper, connection, target):
        from sqlalchemy.exc import IntegrityError

        if isinstance(target, sa_entity):
            # TODO: How to use `session.query(SqlExperiment)` here?
            stmt = mapper.selectable.select()
            for attribute_name in attribute_names:
                stmt = stmt.filter(getattr(sa_entity, attribute_name) == getattr(target, attribute_name))
            stmt = stmt.compile(bind=connection.engine)
            results = connection.execute(stmt)
            if results.rowcount > 0:
                raise IntegrityError(
                    statement=stmt,
                    params=[],
                    orig=Exception(
                        f"DuplicateKeyException in table '{target.__tablename__}' " f"on constraint '{constraint_name}'"
                    ),
                )

    return check_uniqueness


def polyfill_refresh_after_dml(session):
    """
    Run `REFRESH TABLE <tablename>` after each INSERT, UPDATE, and DELETE operation.

    CrateDB is eventually consistent, i.e. write operations are not flushed to
    disk immediately, so readers may see stale data. In a traditional OLTP-like
    application, this is not applicable.

    This SQLAlchemy extension makes sure that data is synchronized after each
    operation manipulating data.

    > `after_{insert,update,delete}` events only apply to the session flush operation
    > and do not apply to the ORM DML operations described at ORM-Enabled INSERT,
    > UPDATE, and DELETE statements. To intercept ORM DML events, use
    > `SessionEvents.do_orm_execute().`
    > -- https://docs.sqlalchemy.org/en/20/orm/events.html#sqlalchemy.orm.MapperEvents.after_insert

    > Intercept statement executions that occur on behalf of an ORM Session object.
    > -- https://docs.sqlalchemy.org/en/20/orm/events.html#sqlalchemy.orm.SessionEvents.do_orm_execute

    > Execute after flush has completed, but before commit has been called.
    > -- https://docs.sqlalchemy.org/en/20/orm/events.html#sqlalchemy.orm.SessionEvents.after_flush

    TODO: Submit patch to `crate-python`, to be enabled by a
          dialect parameter `crate_dml_refresh` or such.
    """  # noqa: E501
    listen(session, "after_flush", do_flush)


def do_flush(session, flush_context):
    """
    SQLAlchemy event handler for the 'after_flush' event,
    invoking `REFRESH TABLE` on each table which has been modified.
    """
    dirty_entities = itertools.chain(session.new, session.dirty, session.deleted)
    dirty_classes = {entity.__class__ for entity in dirty_entities}
    for class_ in dirty_classes:
        refresh_table(session, class_)


def refresh_table(connection, target):
    """
    Invoke a `REFRESH TABLE` statement.
    """
    sql = f"REFRESH TABLE {target.__tablename__}"
    connection.execute(sa.text(sql))



from cratedb_toolkit.util.croud import CroudCall, CroudWrapper


class CloudCluster:
    """
    A wrapper around the CrateDB Cloud cluster API through the `croud` package.
    """

    def __init__(self, cluster_id: str):
        self.cluster_id = cluster_id

    def get_info(self):
        """
        Get cluster information.

        ctk cluster info
        croud clusters get e1e38d92-a650-48f1-8a70-8133f2d5c400 --format=json
        """
        from croud.clusters.commands import clusters_get
        from croud.parser import Argument

        call = CroudCall(
            fun=clusters_get,
            specs=[
                Argument(
                    "id",
                    type=str,
                    help="The ID of the cluster.",
                )
            ],
            arguments=[
                self.cluster_id,
            ],
        )

        wr = CroudWrapper(call=call)
        return wr.invoke()



import logging

import click
from click_aliases import ClickAliasedGroup

from cratedb_toolkit.api.main import ClusterBase, ManagedCluster, StandaloneCluster
from cratedb_toolkit.model import DatabaseAddress, InputOutputResource, TableAddress
from cratedb_toolkit.util.cli import boot_click, make_command

logger = logging.getLogger(__name__)


@click.group(cls=ClickAliasedGroup)  # type: ignore[arg-type]
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    """
    Load data into CrateDB.
    """
    return boot_click(ctx, verbose, debug)


@make_command(cli, name="table")
@click.argument("url")
@click.option(
    "--cluster-id", envvar="CRATEDB_CLOUD_CLUSTER_ID", type=str, required=False, help="CrateDB Cloud cluster identifier"
)
@click.option(
    "--cratedb-sqlalchemy-url", envvar="CRATEDB_SQLALCHEMY_URL", type=str, required=False, help="CrateDB SQLAlchemy URL"
)
@click.option("--cratedb-http-url", envvar="CRATEDB_HTTP_URL", type=str, required=False, help="CrateDB HTTP URL")
@click.option("--schema", envvar="CRATEDB_SCHEMA", type=str, required=False, help="Schema where to import the data")
@click.option("--table", envvar="CRATEDB_TABLE", type=str, required=False, help="Table where to import the data")
@click.option("--format", "format_", type=str, required=False, help="File format of the import resource")
@click.option("--compression", type=str, required=False, help="Compression format of the import resource")
@click.pass_context
def load_table(
    ctx: click.Context,
    url: str,
    cluster_id: str,
    cratedb_sqlalchemy_url: str,
    cratedb_http_url: str,
    schema: str,
    table: str,
    format_: str,
    compression: str,
):
    """
    Import data into CrateDB and CrateDB Cloud clusters.
    """

    error_message = (
        "Either CrateDB Cloud Cluster identifier or CrateDB SQLAlchemy or HTTP URL needs to be supplied. "
        "Use --cluster-id / --cratedb-sqlalchemy-url / --cratedb-http-url CLI options "
        "or CRATEDB_CLOUD_CLUSTER_ID / CRATEDB_SQLALCHEMY_URL / CRATEDB_HTTP_URL environment variables."
    )

    if not cluster_id and not cratedb_sqlalchemy_url and not cratedb_http_url:
        raise KeyError(error_message)

    # When SQLAlchemy URL is not given, but HTTP URL is, compute the former on demand.
    if cluster_id:
        address = None
    elif cratedb_sqlalchemy_url:
        address = DatabaseAddress.from_string(cratedb_sqlalchemy_url)
    elif cratedb_http_url:
        address = DatabaseAddress.from_httpuri(cratedb_sqlalchemy_url)
    else:
        raise KeyError(error_message)

    # Encapsulate source and target parameters.
    resource = InputOutputResource(url=url, format=format_, compression=compression)
    target = TableAddress(schema=schema, table=table)

    # Dispatch "load table" operation.
    cluster: ClusterBase
    if cluster_id:
        cluster = ManagedCluster(cloud_id=cluster_id)
    elif address:
        cluster = StandaloneCluster(address=address)
    else:
        raise NotImplementedError("Unable to select backend")
    return cluster.load_table(resource=resource, target=target)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import typing as t

from azure.storage.blob import BlobServiceClient, ContainerClient
from testcontainers.azurite import AzuriteContainer

from cratedb_toolkit.testing.testcontainers.util import ExtendedDockerContainer


class ExtendedAzuriteContainer(ExtendedDockerContainer, AzuriteContainer):
    """
    An extended Testcontainer for Azurite, emulating Microsoft Azure Blob Storage.

    Features
    - Use the `latest` OCI image from https://quay.io/.
    - Provide convenience methods for getting the Docker-internal endpoint address.

    References
    - https://en.wikipedia.org/wiki/Object_storage
    - https://learn.microsoft.com/en-us/azure/storage/common/storage-use-azurite
    - https://learn.microsoft.com/en-us/azure/storage/blobs/use-azurite-to-run-automated-tests
    - https://github.com/azure/azurite
    """

    def get_real_host_address(self) -> str:
        """
        Provide Docker-internal full endpoint address `<host>:<port>` of the service.
        For example, `172.17.0.4:10000`.
        """
        return f"{self.get_real_host_ip()}:{self._BLOB_SERVICE_PORT}"

    def get_container_endpoint(self, container_name: str):
        container = self.get_container(container_name)
        hostname = self.get_real_host_address()
        return container._format_url(hostname=hostname)

    def get_blob_service_client(self) -> BlobServiceClient:
        """
        Client handle to communicate with the service.
        """
        connection_string = self.get_connection_string()
        return BlobServiceClient.from_connection_string(connection_string, api_version="2019-12-12")

    def create_container(self, container_name: str) -> ContainerClient:
        """
        Create a blob container, and return container client.
        """
        blob_service_client = self.get_blob_service_client()
        return blob_service_client.create_container(container_name)

    def get_container(self, container_name: str) -> ContainerClient:
        """
        Get a blob container, and return container client.
        """
        blob_service_client = self.get_blob_service_client()
        return blob_service_client.get_container_client(container_name)

    def list_blob_names(self, container_name: str) -> t.List[str]:
        """
        Return list of blob names within given container.
        """
        container = self.get_container(container_name)
        return list(container.list_blob_names())



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
"""
A retention policy implementation which reallocates data to "warm" partitions.

It is derived from a corresponding tutorial based on Apache Airflow.
https://community.crate.io/t/cratedb-and-apache-airflow-building-a-hot-cold-storage-data-retention-policy/934

Prerequisites
-------------
- CrateDB 5.2.0 or later
- Tables for storing retention policies need to be created once manually in
  CrateDB. See the file setup/schema.sql in this repository.
"""
import dataclasses
import logging

from cratedb_toolkit.retention.model import RetentionTask

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ReallocateRetentionTask(RetentionTask):
    """
    Represent a data retention task, using the `reallocate` strategy.
    """

    def to_sql(self):
        """
        Render as SQL statement.
        """
        # FIXME: S608 Possible SQL injection vector through string-based query construction
        sql = f"""
        ALTER TABLE {self.table_fullname} PARTITION ({self.partition_column} = {self.partition_value})
        SET ("routing.allocation.require.{self.reallocation_attribute_name}" = '{self.reallocation_attribute_value}');
        """  # noqa: S608
        return sql


class ReallocateRetentionJob:
    """
    Represent a data retention job, using the `reallocate` strategy.
    """

    SQL = """
WITH partitions AS (
  {policy_dql}
),
partition_allocations AS (
  SELECT DISTINCT s.schema_name AS table_schema,
                  s.table_name,
                  s.partition_ident,
                  n.attributes
  FROM sys.shards s
  JOIN sys.nodes n ON s.node['id'] = n.id
)
SELECT *
FROM partitions p
JOIN partition_allocations a ON a.table_schema = p.table_schema
  AND a.table_name = p.table_name
  AND p.partition_ident = a.partition_ident
  AND attributes['{reallocation_attribute_name}'] <> '{reallocation_attribute_value}'
    """



import cratedb_toolkit.datasets.kaggle  # noqa: F401
import cratedb_toolkit.datasets.tutorial  # noqa: F401

from .model import Dataset
from .store import registry


def load_dataset(name: str) -> Dataset:
    return registry.find(name)



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import logging
import sys
import typing as t

import click
from click import ClickException
from click_aliases import ClickAliasedGroup

from cratedb_toolkit.model import DatabaseAddress
from cratedb_toolkit.retention.core import RetentionJob
from cratedb_toolkit.retention.model import JobSettings, RetentionPolicy, RetentionStrategy
from cratedb_toolkit.retention.setup.schema import setup_schema
from cratedb_toolkit.retention.store import RetentionPolicyStore
from cratedb_toolkit.util.cli import (
    boot_click,
    click_options_from_dataclass,
    make_command,
    split_list,
)
from cratedb_toolkit.util.data import jd

logger = logging.getLogger(__name__)


def help_setup():
    """
    Setup database schema for storing retention policies.

    Synopsis
    ========

    # Set up the retention policy database table schema.
    cratedb-retention setup "crate://localhost/"

    """  # noqa: E501


def help_list_policies():
    """
    List data retention policies.

    Examples
    ========

    cratedb-retention list-policies "crate://localhost/"

    """  # noqa: E501


def help_list_tags():
    """
    List all tags.

    Examples
    ========

    cratedb-retention list-tags "crate://localhost/"

    """  # noqa: E501


def help_create_policy():
    """
    Create a data retention policy.

    Examples
    ========

    # Create data retention policy using the DELETE strategy.
    cratedb-retention create-policy --strategy=delete \\
      --table-schema=doc --table-name=raw_metrics \\
      --partition-column=ts_day --retention-period=1 \\
      "crate://localhost/"

    """  # noqa: E501


def help_delete_policy():
    """
    Delete a data retention policy.

    Examples
    ========

    # Delete data retention policy by identifier.
    cratedb-retention delete-policy --id=2af93f28-b315-4bb0-b870-38d3416277f7 \\
      "crate://localhost/"

    """  # noqa: E501


def help_run():
    """
    Invoke data retention workflow.

    Synopsis
    ========

    # Invoke data retention workflow using the `delete` strategy.
    cratedb-retention run --cutoff-day=2023-06-27 --strategy=delete "crate://localhost/"

    """  # noqa: E501


dryrun_option: t.Any = click.option(
    "--dry-run", type=bool, is_flag=True, default=False, help="Do not invoke SQL statements, but display them instead"
)
schema_option: t.Any = click.option(
    "--schema",
    type=str,
    required=False,
    envvar="CRATEDB_EXT_SCHEMA",
    help="Select schema where extension tables are created",
)

cutoff_date_option: t.Any = click.option(
    "--cutoff-day",
    type=str,
    required=False,
    help="Select cutoff date. Retire all data older than »cutoff date« minus "
    "»data retention duration«. The default value is 'today()'.",
)
strategy_option: t.Any = click.option(
    "--strategy",
    type=str,
    required=True,
    help="Which kind of retention strategy to apply",
)
identifier_option: t.Any = click.option(
    "--identifier", "--id", type=str, required=False, help="Identifier of retention policy"
)
tags_option: t.Any = click.option(
    "--tags", type=str, required=False, help="Tags for retention policy, used for grouping and filtering"
)


@click.group(cls=ClickAliasedGroup)  # type: ignore[arg-type]
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    return boot_click(ctx, verbose, debug)


@make_command(cli, "setup", help_setup)
@click.argument("dburi", envvar="CRATEDB_URI")
@dryrun_option
@schema_option
@click.pass_context
def setup(ctx: click.Context, dburi: str, dry_run: bool, schema: t.Optional[str]):
    if not dburi:
        logger.error("Unable to operate without database")
        sys.exit(1)

    # Create `JobSettings` instance, and configure schema name.
    # It is the single source of truth about configuration and runtime settings.
    settings = JobSettings(database=DatabaseAddress.from_string(dburi), dry_run=dry_run)
    if schema is not None:
        settings.policy_table.schema = schema

    # Install database schema.
    setup_schema(settings=settings)


@make_command(cli, "list-policies", help_list_policies, aliases=["list"])
@click.argument("dburi", envvar="CRATEDB_URI")
@schema_option
@click.pass_context
def list_policies(ctx: click.Context, dburi: str, schema: str):
    # Sanity checks.
    if not dburi:
        logger.error("Unable to operate without database")
        sys.exit(1)

    # Create `JobSettings` instance, and configure schema name.
    # It is the single source of truth about configuration and runtime settings.
    settings = JobSettings(
        database=DatabaseAddress.from_string(dburi),
    )
    if schema is not None:
        settings.policy_table.schema = schema

    # Set up adapter to retention policy store.
    store = RetentionPolicyStore(settings=settings)
    jd(store.retrieve())


@make_command(cli, "list-tags", help_list_tags)
@click.argument("dburi", envvar="CRATEDB_URI")
@schema_option
@click.pass_context
def list_tags(ctx: click.Context, dburi: str, schema: str):
    # Sanity checks.
    if not dburi:
        logger.error("Unable to operate without database")
        sys.exit(1)

    # Create `JobSettings` instance, and configure schema name.
    # It is the single source of truth about configuration and runtime settings.
    settings = JobSettings(
        database=DatabaseAddress.from_string(dburi),
    )
    if schema is not None:
        settings.policy_table.schema = schema

    # Set up adapter to retention policy store.
    store = RetentionPolicyStore(settings=settings)
    jd(store.retrieve_tags())


@make_command(cli, "create-policy", help_create_policy, aliases=["create", "add"])
@click.argument("dburi", envvar="CRATEDB_URI")
@schema_option
@click_options_from_dataclass(RetentionPolicy)
@click.pass_context
def create_policy(ctx: click.Context, dburi: str, schema: str, **kwargs):
    # Sanity checks.
    if "strategy" not in kwargs or kwargs["strategy"] is None:
        raise ValueError("--strategy option is required")

    # TODO: Converge to Click converters.
    kwargs["strategy"] = RetentionStrategy(kwargs["strategy"].upper())
    kwargs["tags"] = split_list(kwargs["tags"])

    # Sanity checks.
    if not dburi:
        logger.error("Unable to operate without database")
        sys.exit(1)

    # Create `JobSettings` instance, and configure schema name.
    # It is the single source of truth about configuration and runtime settings.
    settings = JobSettings(
        database=DatabaseAddress.from_string(dburi),
    )
    if schema is not None:
        settings.policy_table.schema = schema

    # Set up adapter to retention policy store.
    store = RetentionPolicyStore(settings=settings)

    # Create a retention policy record.
    policy = RetentionPolicy(**kwargs)
    identifier = store.create(policy)
    logger.info(f"Created new retention policy: {identifier}")


@make_command(cli, "delete-policy", help_delete_policy, aliases=["del", "rm"])
@click.argument("dburi", envvar="CRATEDB_URI")
@schema_option
@identifier_option
@tags_option
@click.pass_context
def delete_policy(ctx: click.Context, dburi: str, schema: str, identifier: str, tags: str):
    # TODO: Converge to Click converters.
    tag_list = set(split_list(tags))

    # Sanity checks.
    if not dburi:
        logger.error("Unable to operate without database")
        sys.exit(1)

    # Create `JobSettings` instance, and configure schema name.
    # It is the single source of truth about configuration and runtime settings.
    settings = JobSettings(
        database=DatabaseAddress.from_string(dburi),
    )
    if schema is not None:
        settings.policy_table.schema = schema

    # Set up adapter to retention policy store.
    store = RetentionPolicyStore(settings=settings)

    # Create a retention policy record.
    if identifier:
        store.delete(identifier=identifier)
    elif tags:
        store.delete_by_all_tags(tags=tag_list)
    else:
        message = "Unable to obtain deletion constraint"
        logger.error(message)
        raise ClickException(message)


@make_command(cli, "run", help_run)
@click.argument("dburi", envvar="CRATEDB_URI")
@dryrun_option
@schema_option
@cutoff_date_option
@strategy_option
@tags_option
@click.pass_context
def run(
    ctx: click.Context, dburi: str, cutoff_day: str, strategy: str, tags: str, dry_run: bool, schema: t.Optional[str]
):
    # TODO: Converge to Click converters.
    strategy_type = RetentionStrategy(strategy.upper())
    tag_list = set(split_list(tags))

    if not dburi:
        logger.error("Unable to operate without database")
        sys.exit(1)

    # Create `JobSettings` instance.
    # It is the single source of truth about configuration and runtime settings.
    settings = JobSettings(
        database=DatabaseAddress.from_string(dburi),
        strategy=strategy_type,
        tags=tag_list,
        cutoff_day=cutoff_day,
        dry_run=dry_run,
    )
    if schema is not None:
        settings.policy_table.schema = schema

    # Invoke the data retention job.
    job = RetentionJob(settings=settings)
    job.start()



try:
    from importlib.metadata import PackageNotFoundError, version
except (ImportError, ModuleNotFoundError):  # pragma:nocover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore[assignment,no-redef,unused-ignore]

__appname__ = "cratedb-toolkit"

try:
    __version__ = version(__appname__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



import abc
import dataclasses
import json
import logging
import typing as t
from abc import abstractmethod

from cratedb_toolkit.api.guide import GuidingTexts
from cratedb_toolkit.cluster.util import get_cluster_info
from cratedb_toolkit.exception import CroudException, OperationFailed
from cratedb_toolkit.model import ClusterInformation, DatabaseAddress, InputOutputResource, TableAddress

logger = logging.getLogger(__name__)


class ClusterBase(abc.ABC):
    @abstractmethod
    def load_table(self, resource: InputOutputResource, target: TableAddress):
        raise NotImplementedError("Child class needs to implement this method")


@dataclasses.dataclass
class ManagedCluster(ClusterBase):
    """
    Wrap a managed CrateDB database cluster on CrateDB Cloud.
    """

    cloud_id: str
    address: t.Optional[DatabaseAddress] = None
    info: t.Optional[ClusterInformation] = None

    def __post_init__(self):
        logger.info(f"Connecting to CrateDB Cloud Cluster: {self.cloud_id}")

    def load_table(self, resource: InputOutputResource, target: t.Optional[TableAddress] = None):
        """
        Load data into a database table on CrateDB Cloud.

        Synopsis
        --------
        export CRATEDB_CLOUD_CLUSTER_ID=95998958-4d96-46eb-a77a-a894e7dde128
        ctk load table https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/data_weather.csv.gz

        https://console.cratedb.cloud
        """
        from cratedb_toolkit.io.croud import CloudIo

        target = target or TableAddress()

        try:
            cluster_info = get_cluster_info(cluster_id=self.cloud_id)
            logger.info(
                f"Cluster information: name={cluster_info.cloud.get('name')}, url={cluster_info.cloud.get('url')}"
            )
            cio = CloudIo(cluster_id=self.cloud_id)
        except CroudException as ex:
            msg = f"Connecting to cluster resource failed: {self.cloud_id}. Reason: {ex}"
            if "Resource not found" in str(ex):
                logger.error(msg)
                return None, False
            logger.exception(msg)
            raise OperationFailed(msg) from ex

        try:
            job_info, success = cio.load_resource(resource=resource, target=target)
            logger.info("Job information:\n%s", json.dumps(job_info, indent=2))
            # TODO: Explicitly report about `failed_records`, etc.
            texts = GuidingTexts(
                admin_url=cluster_info.cloud["url"],
                table_name=job_info["destination"]["table"],
            )
            if success:
                logger.info("Data loading was successful: %s", texts.success())
                return job_info, success
            else:
                # TODO: Add "reason" to exception message.
                logger.error(f"Data loading failed: {texts.error()}")
                raise OperationFailed("Data loading failed")

            # When exiting so, it is expected that error logging has taken place appropriately.
        except CroudException as ex:
            msg = "Data loading failed: Unknown error"
            logger.exception(msg)
            raise OperationFailed(msg) from ex


@dataclasses.dataclass
class StandaloneCluster(ClusterBase):
    """
    Wrap a standalone CrateDB database cluster.
    """

    address: DatabaseAddress
    info: t.Optional[ClusterInformation] = None

    def load_table(self, resource: InputOutputResource, target: TableAddress):
        """
        Load data into a database table on a standalone CrateDB Server.

        Synopsis
        --------
        export CRATEDB_SQLALCHEMY_URL=crate://crate@localhost:4200/testdrive/demo

        ctk load table influxdb2://example:token@localhost:8086/testdrive/demo
        ctk load table mongodb://localhost:27017/testdrive/demo
        """
        source_url = resource.url
        target_url = self.address.dburi
        if source_url.startswith("influxdb"):
            from cratedb_toolkit.io.influxdb import influxdb_copy

            source_url = source_url.replace("influxdb2://", "http://")
            if not influxdb_copy(source_url, target_url, progress=True):
                msg = "Data loading failed"
                logger.error(msg)
                raise OperationFailed(msg)
        elif source_url.startswith("mongodb"):
            from cratedb_toolkit.io.mongodb.api import mongodb_copy

            if not mongodb_copy(source_url, target_url, progress=True):
                msg = "Data loading failed"
                logger.error(msg)
                raise OperationFailed(msg)
        else:
            raise NotImplementedError("Importing resource not implemented yet")



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import dataclasses
import logging
import sys
import textwrap
import typing as t

import click

from cratedb_toolkit.util.common import setup_logging

logger = logging.getLogger(__name__)


def boot_click(ctx: click.Context, verbose: bool = False, debug: bool = False):
    """
    Bootstrap the CLI application.
    """

    # Adjust log level according to `verbose` / `debug` flags.
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG

    # Setup logging, according to `verbose` / `debug` flags.
    setup_logging(level=log_level, verbose=verbose)


def split_list(value: str, delimiter: str = ",") -> t.List[str]:
    if value is None:
        return []
    return [c.strip() for c in value.split(delimiter)]


def to_list(x: t.Any, default: t.List[t.Any] = None) -> t.List[t.Any]:
    if not isinstance(default, t.List):
        raise ValueError("Default value is not a list")
    if x is None:
        return default
    if not isinstance(x, t.Iterable) or isinstance(x, str):
        return [x]
    elif isinstance(x, list):
        return x
    else:
        return list(x)


def docstring_format_verbatim(text: t.Optional[str]) -> str:
    """
    Format docstring to be displayed verbatim as a help text by Click.

    - https://click.palletsprojects.com/en/8.1.x/documentation/#preventing-rewrapping
    - https://github.com/pallets/click/issues/56
    """
    text = text or ""
    text = textwrap.dedent(text)
    lines = [line if line.strip() else "\b" for line in text.splitlines()]
    return "\n".join(lines)


def boot_with_dburi():
    """
    Obtain a single positional argument from the command line,
    the database URI, in SQLAlchemy-compatible string format.
    """
    setup_logging()
    try:
        dburi = sys.argv[1]
    except IndexError:
        dburi = "crate://localhost/"
    return dburi


def make_command(cli, name, helpfun=None, aliases=None):
    """
    Convenience shortcut for creating a subcommand.
    """
    kwargs = {}
    if helpfun:
        kwargs["help"] = docstring_format_verbatim(helpfun.__doc__)
    return cli.command(
        name,
        context_settings={"max_content_width": 120},
        aliases=aliases,
        **kwargs,
    )


def click_options_from_dataclass(cls):
    """
    Generate Click command line options dynamically from dataclass definition.

    Inspired by:
    https://www.zonca.dev/posts/2022-10-26-click-commandline-class-arguments
    """

    def decorator(f):
        fields = dataclasses.fields(cls)
        for field in fields:
            if field.metadata.get("read_only", False):
                continue
            field_name = field.name.replace("_", "-")
            if field_name not in ["self"]:
                click.option("--" + field_name, required=False, type=str, help=field.metadata.get("help", ""))(f)
        return f

    return decorator



from cratedb_toolkit.datasets.model import Dataset
from cratedb_toolkit.datasets.store import registry

the_weather_dataset_daily = Dataset(
    title="The Weather Dataset: Daily Weather",
    reference="kaggle://guillemservera/global-daily-climate-data/daily_weather.parquet",
    documentation="https://www.kaggle.com/datasets/guillemservera/global-daily-climate-data",
    ddl="""
        CREATE TABLE IF NOT EXISTS {table} (
           "station_id" TEXT,
           "city_name" TEXT,
           "date" TIMESTAMP WITHOUT TIME ZONE,
           "season" TEXT,
           "avg_temp_c" REAL,
           "min_temp_c" REAL,
           "max_temp_c" REAL,
           "precipitation_mm" REAL,
           "snow_depth_mm" REAL,
           "avg_wind_dir_deg" REAL,
           "avg_wind_speed_kmh" REAL,
           "peak_wind_gust_kmh" REAL,
           "avg_sea_level_pres_hpa" REAL,
           "sunshine_total_min" REAL,
           INDEX city_name_ft using fulltext (city_name)
        )
        """,
)


the_weather_dataset_cities = Dataset(
    title="The Weather Dataset: Cities",
    reference="kaggle://guillemservera/global-daily-climate-data/cities.csv",
    documentation="https://www.kaggle.com/datasets/guillemservera/global-daily-climate-data",
    ddl="""
        CREATE TABLE IF NOT EXISTS {table} (
           "station_id" TEXT,
           "city_name" TEXT,
           "country" TEXT,
           "state" TEXT,
           "iso2" TEXT,
           "iso3" TEXT,
           "loc" GEO_POINT
        )
    """,
)


registry.add(the_weather_dataset_daily)
registry.add(the_weather_dataset_cities)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

"""
Translate a MongoDB collection schema into a CrateDB CREATE TABLE expression.

Given a generated MongoDB collection schema, this will translate that schema
into a CREATE TABLE statement, mapping fields to columns and the collection
name to the table name.

In the case where there are type conflicts (for example, 40% of the values
for a field are integers, and 60% are strings), the translator will choose
the type with the greatest proportion.
"""

from functools import reduce

TYPES = {
    "DATETIME": "TIMESTAMP WITH TIME ZONE",
    "INT64": "INTEGER",
    "STRING": "TEXT",
    "BOOLEAN": "BOOLEAN",
    "INTEGER": "INTEGER",
    "FLOAT": "FLOAT",
    "ARRAY": "ARRAY",
    "OBJECT": "OBJECT",
}

BASE = """
CREATE TABLE IF NOT EXISTS "{schema}"."{table}" (\n{columns}\n);
"""

COLUMN = '"{column_name}" {type}'

OBJECT = "OBJECT ({object_type}) AS (\n{definition}\n)"


def get_columns_definition(columns):
    columns_definition = []
    for column in columns:
        if column[1]:
            item = f"{column[1]}\n{column[0]}"
        else:
            item = column[0]
        columns_definition.append(item)
    return columns_definition


def translate_object(schema):
    """
    Translate an object field schema definition into a CrateDB dynamic object column.
    """

    columns = []
    object_type = "DYNAMIC"
    for fieldname, field in schema.items():
        sql_type, comment = determine_type(field)
        columns.append((COLUMN.format(column_name=fieldname, type=sql_type), comment))
    columns_definition = get_columns_definition(columns)
    return OBJECT.format(
        object_type=object_type,
        definition=",\n".join(columns_definition),
    )


def translate_array(schema):
    """
    Translate an array field schema definition into a CrateDB array column.
    """

    subtype, comment = determine_type(schema)
    if comment:
        return f"{comment}\nARRAY({subtype})"
    else:
        return f"ARRAY({subtype})"


def determine_type(schema):
    """
    Determine the type of a specific field schema.
    """

    types = schema.get("types", {})
    type_ = max(types, key=lambda item: types[item]["count"])
    if type_ in TYPES:
        sql_type = TYPES.get(type_)
        if sql_type == "OBJECT":
            sql_type = translate_object(types["OBJECT"]["document"])
        elif sql_type == "ARRAY":
            sql_type = translate_array(types["ARRAY"])

        if len(types) > 1:
            return (sql_type, proportion_string(types))
        return (sql_type, None)
    return ("UNKNOWN", None)


def proportion_string(types: dict) -> str:
    """
    Convert a list of types into a string explaining the proportions of each type.
    """

    total = reduce(lambda x, y: x + types[y]["count"], list(types.keys()), 0)
    summary = "-- ⬇️ Types: "
    proportions = []
    for type_ in types:
        proportions.append(f"{type_}: {round((types[type_]['count']/total)*100, 2)}%")
    return " " + (summary + ", ".join(proportions))


def indent_sql(query: str) -> str:
    """
    Indent an SQL query based on opening and closing brackets.
    """

    indent = 0
    lines = query.split("\n")
    for idx, line in enumerate(lines):
        lines[idx] = (" " * indent) + line
        if len(line) >= 1:
            if line[-1] == "(":
                indent += 4
            elif line[-1] == ")":
                indent -= 4
    return "\n".join(lines)


def translate(schemas, schemaname: str = None):
    """
    Translate a schema definition for a set of MongoDB collection schemas.

    This results in a set of CrateDB compatible CREATE TABLE expressions
    corresponding to the set of MongoDB collection schemas.
    """
    schemaname = schemaname or "doc"

    tables = list(schemas.keys())
    sql_queries = {}
    for tablename in tables:
        collection = schemas[tablename]
        columns = []
        for fieldname, field in collection["document"].items():
            sql_type, comment = determine_type(field)
            if sql_type != "UNKNOWN":
                columns.append((COLUMN.format(column_name=fieldname, type=sql_type), comment))

        columns_definition = get_columns_definition(columns)
        sql_queries[tablename] = indent_sql(
            BASE.format(schema=schemaname, table=tablename, columns=",\n".join(columns_definition))
        )
    return sql_queries



import os
import sys
import typing as t

from crate.crash.command import main


def run_crash(
    hosts: str, command: str, output_format: str = None, schema: str = None, username: str = None, password: str = None
):
    """
    Run interactive CrateDB database shell, using `crash`.
    """

    cmd = ["crash", "--hosts", hosts]
    if username:
        cmd += ["--username", username]
    if schema:
        cmd += ["--schema", schema]
    if command:
        cmd += ["--command", command]
    if output_format:
        cmd += ["--format", output_format]
    sys.argv = cmd
    if password:
        os.environ["CRATEPW"] = password
    main()


def get_crash_output_formats() -> t.List[str]:
    """
    Inquire the output formats `crash` understands.
    """
    from crate.crash.outputs import OutputWriter
    from crate.crash.printer import PrintWrapper

    return OutputWriter(PrintWrapper(), False).formats



import sys

import click
from click_aliases import ClickAliasedGroup

from cratedb_toolkit.cluster.util import get_cluster_info
from cratedb_toolkit.exception import CroudException
from cratedb_toolkit.util import jd
from cratedb_toolkit.util.cli import boot_click, make_command


@click.group(cls=ClickAliasedGroup)  # type: ignore[arg-type]
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    """
    Run cluster operations on CrateDB and CrateDB Cloud.
    """
    return boot_click(ctx, verbose, debug)


@make_command(cli, name="info")
@click.option(
    "--cluster-id", envvar="CRATEDB_CLOUD_CLUSTER_ID", type=str, required=True, help="CrateDB Cloud cluster identifier"
)
@click.pass_context
def info(ctx: click.Context, cluster_id: str):
    """
    Display CrateDB Cloud Cluster information

    ctk cluster info
    ctk cluster info --cluster-id=e1e38d92-a650-48f1-8a70-8133f2d5c400
    croud clusters get e1e38d92-a650-48f1-8a70-8133f2d5c400 --format=json
    """
    cluster_info = get_cluster_info(cluster_id=cluster_id)
    try:
        jd(cluster_info.asdict())

    # When exiting so, it is expected that error logging has taken place appropriately.
    except CroudException:
        sys.exit(1)



# Copyright (c) 2021-2023, Crate.io Inc.
# Distributed under the terms of the AGPLv3 license, see LICENSE.
import abc
import logging
import typing as t
import uuid

import sqlalchemy as sa
from sqlalchemy import MetaData, Table
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm import Session
from sqlalchemy.sql.selectable import NamedFromClause

from cratedb_toolkit.retention.model import JobSettings, RetentionPolicy, RetentionStrategy
from cratedb_toolkit.util.database import DatabaseAdapter, sa_is_empty

logger = logging.getLogger(__name__)


class SQLAlchemyTagHelperMixin:
    """
    Support tags stored within an `OBJECT(DYNAMIC)` column in CrateDB, managed with SQLAlchemy.
    """

    tag_column = "tags"

    @abc.abstractmethod
    def execute(self, statement):
        pass

    @abc.abstractmethod
    def synchronize(self):
        pass

    def get_tags_constraints(self, tags: t.Union[t.List[str], t.Set[str]]):
        """
        Return list of SQL WHERE constraint clauses from given tags.
        """
        table: NamedFromClause = self.table  # type: ignore[attr-defined]
        constraints = []
        for tag in tags:
            if not tag:
                continue
            constraint = table.c[self.tag_column][tag] != sa.Null()
            constraints.append(constraint)
        return sa.and_(sa.true(), *constraints)

    def tags_exist(self, tags: t.Union[t.List[str], t.Set[str]]):
        """
        Check if given tags exist in the database.

        When not, no query can yield results, so we do not need to bother about
        failing JOIN operations because of non-existing tags.

        TODO: Create corresponding issue at crate/crate.
        """
        table = self.table  # type: ignore[attr-defined]
        where_clause = self.get_tags_constraints(tags)
        if sa_is_empty(where_clause):
            return False
        selectable = sa.select(table).where(where_clause)
        try:
            self.execute(selectable)
        except ProgrammingError as ex:
            if "ColumnUnknownException" in str(ex):
                return False
            else:
                raise
        return True

    def delete_by_tag(self, tag: str):
        """
        Delete retention policy by tag.
        """
        return self.delete_by_all_tags([tag])

    def delete_by_all_tags(self, tags: t.Union[t.List[str], t.Set[str]]):
        """
        Delete retention policy by tag.
        """
        if not tags:
            logger.warning("No tags obtained, skipping deletion")
            return 0

        if not self.tags_exist(tags):
            logger.warning(f"No retention policies found with tags: {tags}")
            return 0

        table = self.table  # type: ignore[attr-defined]
        where_clause = self.get_tags_constraints(tags)
        if sa_is_empty(where_clause):
            logger.warning("Unable to compute criteria for deletion")
            return 0
        deletable = sa.delete(table).where(where_clause)  # type: ignore[arg-type]
        result = self.execute(deletable)
        self.synchronize()
        return result.rowcount


class RetentionPolicyStore(SQLAlchemyTagHelperMixin):
    """
    A wrapper around the retention policy database table.
    """

    def __init__(self, settings: JobSettings):
        self.settings = settings

        logger.info(
            f"Connecting to database {self.settings.database.safe}, " f"table {self.settings.policy_table.fullname}"
        )

        if self.settings.policy_table.table is None:
            raise ValueError("Unable to create RetentionPolicyStore without table name")

        # Set up generic database adapter.
        self.database: DatabaseAdapter = DatabaseAdapter(dburi=self.settings.database.dburi)

        # Set up SQLAlchemy Core adapter for retention policy table.
        metadata = MetaData(schema=self.settings.policy_table.schema)
        self.table = Table(self.settings.policy_table.table, metadata, autoload_with=self.database.engine)

    def create(self, policy: RetentionPolicy, ignore: t.Optional[str] = None):
        """
        Create a new retention policy, and return its identifier.
        """

        ignore = ignore or ""

        # Sanity checks.
        if policy.table_schema is None:
            raise ValueError("Table schema needs to be defined")
        if policy.table_name is None:
            raise ValueError("Table name needs to be defined")
        if self.exists(policy):
            if not ignore.startswith("DuplicateKey"):
                raise ValueError(
                    f"Retention policy for table '{policy.table_schema}.{policy.table_name}' already exists"
                )

        table = self.table
        # TODO: Add UUID as converter to CrateDB driver?
        identifier = str(uuid.uuid4())
        data = policy.to_storage_dict(identifier=identifier)
        insertable = sa.insert(table).values(**data).returning(table.c.id)
        cursor = self.execute(insertable)
        identifier = cursor.one()[0]
        self.synchronize()
        return identifier

    def retrieve(self):
        """
        Retrieve all records from database table.

        TODO: Add filtering capabilities.
        """

        # Run SELECT statement, and return result.
        selectable = sa.select(self.table)
        records = self.query(selectable)
        records = list(map(self.row_to_record, records))
        return records

    def retrieve_policies(self, strategy: RetentionStrategy, tags):
        """
        Retrieve effective policies to process, by strategy and tags.
        """
        table = self.table
        selectable = sa.select(table).where(table.c.strategy == strategy.to_database())
        criteria = self.get_tags_constraints(tags)
        selectable = selectable.where(criteria)
        for row in self.query(selectable):
            policy = RetentionPolicy.from_record(row)
            yield policy

    def retrieve_tags(self):
        """
        Retrieve all tags from database table.

        TODO: Add filtering capabilities.
        """

        # Run SELECT statement, and return result.
        selectable = sa.select(self.table.c.tags)
        records = self.query(selectable)
        tags = []
        for record in records:
            tags += self.row_to_record(record)["tags"]
        tags = sorted(set(tags))
        return tags

    @staticmethod
    def row_to_record(item):
        """
        Compute serializable representation from database row.

        Convert between representation of tags. In Python land, it's a `set`,
        while in database land, it's an `OBJECT`.
        """
        if isinstance(item["tags"], dict):
            item["tags"] = sorted(item["tags"].keys())
        return item

    def delete(self, identifier: str):
        """
        Delete retention policy by identifier.
        """
        table = self.table
        constraint = table.c.id == identifier
        deletable = sa.delete(table).where(constraint)
        result = self.execute(deletable)
        self.synchronize()
        if result.rowcount == 0:
            logger.warning(f"Retention policy not found with id: {identifier}")
        return result.rowcount

    def execute(self, statement):
        """
        Execute SQL statement, and return result object.
        """
        with Session(self.database.engine) as session:
            result = session.execute(statement)
            session.commit()
            return result

    def query(self, statement):
        """
        Execute SQL statement, fetch result rows, and return them converted to dictionaries.
        """
        cursor = self.execute(statement)
        rows = cursor.mappings().fetchall()
        records = [dict(row.items()) for row in rows]
        return records

    def exists(self, policy: RetentionPolicy):
        """
        Check if retention policy for specific table already exists.
        """
        table = self.table
        selectable = sa.select(table).where(
            table.c.table_schema == policy.table_schema,
            table.c.table_name == policy.table_name,
        )
        result = self.query(selectable)
        return bool(result)

    def synchronize(self):
        """
        Synchronize data by issuing `REFRESH TABLE` statement.
        """
        sql = f"REFRESH TABLE {self.settings.policy_table.fullname};"
        self.database.run_sql(sql)



import typing as t

import sqlalchemy as sa


def patch_inspector():
    """
    When using `get_table_names()`, make sure the correct schema name gets used.

    Apparently, SQLAlchemy does not honor the `search_path` of the engine, when
    using the inspector?

    FIXME: Bug in CrateDB SQLAlchemy dialect?
    """

    def get_effective_schema(engine: sa.Engine):
        schema_name_raw = engine.url.query.get("schema")
        schema_name = None
        if isinstance(schema_name_raw, str):
            schema_name = schema_name_raw
        elif isinstance(schema_name_raw, tuple):
            schema_name = schema_name_raw[0]
        return schema_name

    from crate.client.sqlalchemy.dialect import CrateDialect

    get_table_names_dist = CrateDialect.get_table_names

    def get_table_names(self, connection: sa.Connection, schema: t.Optional[str] = None, **kw: t.Any) -> t.List[str]:
        if schema is None:
            schema = get_effective_schema(connection.engine)
        return get_table_names_dist(self, connection=connection, schema=schema, **kw)

    CrateDialect.get_table_names = get_table_names  # type: ignore



from cratedb_toolkit.datasets.model import Dataset
from cratedb_toolkit.datasets.store import registry

devices_info = Dataset(
    reference="tutorial/devices-info",
    data_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/devices_info.json.gz",
    init_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/load_devices_info.sql",
    init_includes_loading=True,
)

devices_readings = Dataset(
    reference="tutorial/devices-readings",
    data_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/devices_readings.json.gz",
    init_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/load_devices_readings.sql",
    init_includes_loading=True,
)

marketing = Dataset(
    reference="tutorial/marketing",
    data_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/data_marketing.json.gz",
    init_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/load_marketing.sql",
    init_includes_loading=True,
)

netflix = Dataset(
    reference="tutorial/netflix",
    data_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/data_netflix.json.gz",
    init_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/load_netflix.sql",
    init_includes_loading=True,
)

weather = Dataset(
    reference="tutorial/weather-basic",
    data_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/data_weather.csv.gz",
    init_url="https://github.com/crate/cratedb-datasets/raw/main/cloud-tutorials/load_weather.sql",
    init_includes_loading=True,
)

registry.add(devices_info)
registry.add(devices_readings)
registry.add(marketing)
registry.add(netflix)
registry.add(weather)



import json
import sys
import typing as t


def jd(data: t.Any):
    """
    Pretty-print JSON with indentation.
    """
    print(json.dumps(data, indent=2), file=sys.stdout)  # noqa: T201


def str_contains(haystack, *needles):
    """
    Whether haystack contains any of the provided needles.
    """
    haystack = str(haystack)
    return any(needle in haystack for needle in needles)



import io
import logging
import typing as t

import pymongo
import pymongo.database
import rich
from bson.raw_bson import RawBSONDocument
from rich.syntax import Syntax

from .export import collection_to_json
from .extract import extract_schema_from_collection
from .translate import translate as translate_schema
from .util import parse_input_numbers

logger = logging.getLogger(__name__)


def gather_collections(database) -> t.List[str]:
    """
    Gather a list of collections to use from a MongoDB database, based on user input.
    """

    collections = database.list_collection_names()

    tbl = rich.table.Table(show_header=True, header_style="bold blue")
    tbl.add_column("Id", width=3)
    tbl.add_column("Collection Name")
    tbl.add_column("Estimated Size")

    for i, c in enumerate(collections):
        tbl.add_row(str(i), c, str(database[c].estimated_document_count()))

    rich.print(tbl)

    rich.print("\nCollections to exclude: (eg: '0 1 2', '0, 1, 2', '0-2')")

    collections_to_ignore = parse_input_numbers(input("> "))
    filtered_collections = []
    for i, c in enumerate(collections):
        if i not in collections_to_ignore:
            filtered_collections.append(c)

    # MongoDB 2 does not understand `include_system_collections=False`.
    filtered_collections = [item for item in filtered_collections if not item.startswith("system.")]

    return filtered_collections


def get_mongodb_client_database(args, **kwargs) -> t.Tuple[pymongo.MongoClient, pymongo.database.Database]:
    client: pymongo.MongoClient
    if args.url:
        client = pymongo.MongoClient(args.url, **kwargs)
    else:
        client = pymongo.MongoClient(args.host, int(args.port), **kwargs)
    db: pymongo.database.Database = client.get_database(args.database)
    return client, db


def extract(args) -> t.Dict[str, t.Any]:
    """
    Extract schemas from MongoDB collections.

    This asks the user for which collections they would like to extract,
    iterates over these collections and returns a dictionary of schemas for
    each of the selected collections.
    """
    client, db = get_mongodb_client_database(args)
    if args.collection:
        filtered_collections = [args.collection]
    else:
        filtered_collections = gather_collections(db)

    if not filtered_collections:
        rich.print("\nExcluding all collections. Nothing to do.")
        exit(0)

    if args.scan:
        partial = args.scan == "partial"
    else:
        rich.print("\nDo a [red bold]full[/red bold] collection scan?")
        rich.print("A full scan will iterate over all documents in the collection, a partial only one document. (Y/n)")
        full = input(">  ").strip().lower()

        partial = full != "y"

        rich.print(f"\nExecuting a [red bold]{'partial' if partial else 'full'}[/red bold] scan...")

    schemas = {}
    for collection in filtered_collections:
        schemas[collection] = extract_schema_from_collection(db[collection], partial)
    return schemas


def translate(schemas, schemaname: str = None) -> t.Dict[str, str]:
    """
    Translate a given schema into SQL DDL statements compatible with CrateDB.
    """
    result: t.Dict[str, str] = {}
    sql_queries = translate_schema(schemas=schemas, schemaname=schemaname)
    for collection, query in sql_queries.items():
        result[collection] = query
        syntax = Syntax(query, "sql")
        rich.print(f"Collection [blue bold]'{collection}'[/blue bold]:")
        rich.print(syntax)
        rich.print()
    return result


def export(args) -> t.IO[bytes]:
    """
    Export MongoDB collection into JSON format.

    TODO: Run on multiple collections, like `extract`.
    """
    buffer = io.BytesIO()
    client, db = get_mongodb_client_database(args, document_class=RawBSONDocument)
    collection_to_json(db[args.collection], file=buffer)
    buffer.seek(0)
    return buffer



import argparse
import logging

from cratedb_toolkit.io.mongodb.core import export, extract, translate
from cratedb_toolkit.model import DatabaseAddress
from cratedb_toolkit.util.cr8 import cr8_insert_json
from cratedb_toolkit.util.database import DatabaseAdapter

logger = logging.getLogger(__name__)


def mongodb_copy(source_url, target_url, progress: bool = False):
    """
    Synopsis
    --------
    export CRATEDB_SQLALCHEMY_URL=crate://crate@localhost:4200/testdrive/demo
    ctk load table mongodb://localhost:27017/testdrive/demo

    Backlog
    -------
    TODO: Run on multiple collections.
    TODO: Run on the whole database.
    TODO: Accept parameters like `if_exists="append,replace"`.
    TODO: Propagate parameters like `scan="full"`.
    TODO: Handle timestamp precision(s)?
    """
    logger.info("Running MongoDB copy")

    # Decode database URL.
    mongodb_address = DatabaseAddress.from_string(source_url)
    mongodb_uri, mongodb_collection_address = mongodb_address.decode()
    mongodb_database = mongodb_collection_address.schema
    mongodb_collection = mongodb_collection_address.table

    # 1. Extract schema from MongoDB collection.
    logger.info(f"Extracting schema from MongoDB: {mongodb_database}.{mongodb_collection}")
    extract_args = argparse.Namespace(
        url=str(mongodb_uri), database=mongodb_database, collection=mongodb_collection, scan="full"
    )
    mongodb_schema = extract(extract_args)
    count = mongodb_schema[mongodb_collection]["count"]
    if not count > 0:
        logger.error(f"No results when extracting schema from MongoDB: {mongodb_database}.{mongodb_collection}")
        return False

    # 2. Translate schema to SQL DDL.
    cratedb_address = DatabaseAddress.from_string(target_url)
    cratedb_uri, cratedb_table_address = cratedb_address.decode()
    ddl = translate(mongodb_schema, schemaname=cratedb_table_address.schema)

    # 3. Load schema SQL DDL into CrateDB.
    cratedb = DatabaseAdapter(dburi=str(cratedb_uri))
    for collection, query in ddl.items():
        logger.info(f"Creating table for collection '{collection}': {query}")
        cratedb.run_sql(query)

    # 4. Transfer data to CrateDB.
    """
    migr8 export --host localhost --port 27017 --database test_db --collection test | \
        cr8 insert-json --hosts localhost:4200 --table test
    """
    logger.info(
        f"Transferring data from MongoDB to CrateDB: "
        f"source={mongodb_collection_address.fullname}, target={cratedb_table_address.fullname}"
    )
    export_args = argparse.Namespace(url=str(mongodb_uri), database=mongodb_database, collection=mongodb_collection)
    buffer = export(export_args)
    cr8_insert_json(infile=buffer, hosts=cratedb_address.httpuri, table=cratedb_table_address.fullname)

    return True



import re


def parse_input_numbers(s: str):
    """
    Parse an input string for numbers and ranges.

    Supports strings like '0 1 2', '0, 1, 2' as well as ranges such as
    '0-2'.
    """

    options: list = []
    for option in re.split(", | ", s):
        match = re.search(r"(\d+)-(\d+)", option)
        if match:
            lower, upper = sorted([match.group(1), match.group(2)])
            options = options + list(range(int(lower), int(upper) + 1))
        else:
            try:
                options.append(int(option))
            except ValueError:
                pass
    return options
